

'-------------------------------------------------------------------------------
'Object tracking...
DIM MyImageMatch AS cv_WrapperImageMatch_t      'struct for sending parameters to object tracking DLL
'-------------------------------------------------------------------------------



'-------------------------------------------------------------------------------
' matching of prior image to current image to get global translation and rotation

DIM DiscROI AS QBITMAPEX            'current area of the optic disc, force 32-bit stride
DIM SubROI AS QBITMAPEX             'sub region of current optic disc to search in reference disc, force 32-bit stride

CONST MaxTranslations  = 32                         'check of disc movement relative to reference
DIM DiscTrans(MaxTranslations)  AS PointCoordType   'array holding coordinates of disc translations
DEFINT DiscMatchCrossClr = clPurple
'-------------------------------------------------------------------------------





'-------------------------------------------------------------------------------
'  Break up optic disc reference into small regions of interest (RO)
'  then find matches in the currently loaded optic disc image.
'  These matches should show the changes in disc location during
'  acquisition. However in its current state errors in matching
'  are too large for it to work accurately
'-------------------------------------------------------------------------------

SUB GetOpticDiscMovement

    'no reference set yet
    IF (ReferenceDiscBMP.Width <> ReferenceImageWidth) THEN  EXIT SUB

    'get an area of equivalent size to the reference optic dics, check bounds
    DIM Dest AS QRECT
        Dest.Top = 0
        Dest.Left = 0
        Dest.Bottom = ReferenceDiscBMP.Height
        Dest.Right = ReferenceDiscBMP.Width


    'get current disc ROI centered on **user's selection **
    DIM ObjectAreaRect AS QRECT
        ObjectAreaRect.Top =    DiskCenter.ScreenY - ReferenceDiscBMP.Height\2
        ObjectAreaRect.Bottom = DiskCenter.ScreenY + ReferenceDiscBMP.Height\2
        ObjectAreaRect.Left =   DiskCenter.ScreenX - ReferenceDiscBMP.Width\2
        ObjectAreaRect.Right  = DiskCenter.ScreenX + ReferenceDiscBMP.Width\2

    'check if area is out of bounds
    '-------------------------------------------------------------------
    IF ObjectAreaRect.Right > SLO_Orig.Width THEN 
        ReferenceForceOffsetX = SLO_Orig.Width - ObjectAreaRect.Right
        ObjectAreaRect.Right = SLO_Orig.Width
        ObjectAreaRect.Left = ObjectAreaRect.Left + ReferenceForceOffsetX
    END IF

    IF ObjectAreaRect.Left < 0 THEN 
        ReferenceForceOffsetX = ObjectAreaRect.Left * -1
        ObjectAreaRect.Left = 0
        ObjectAreaRect.Right = ObjectAreaRect.Right + ReferenceForceOffsetX
    END IF

    IF ObjectAreaRect.Bottom > SLO_Orig.Height THEN 
        ReferenceForceOffsetY = SLO_Orig.Height - ObjectAreaRect.Bottom
        ObjectAreaRect.Bottom = SLO_Orig.Height
        ObjectAreaRect.Top = ObjectAreaRect.Top + ReferenceForceOffsetY
    END IF

    IF ObjectAreaRect.Top < 0 THEN 
        ReferenceForceOffsetY = ObjectAreaRect.Top * -1
        ObjectAreaRect.Top = 0
        ObjectAreaRect.Bottom = ObjectAreaRect.Bottom + ReferenceForceOffsetY
    END IF

    IF ObjectAreaRect.Left < 0 THEN Showmessage "disc too off center": EXIT SUB
    IF ObjectAreaRect.Right > SLOimg.Width THEN Showmessage "disc too off center": EXIT SUB
    IF ObjectAreaRect.Top < 0 THEN  Showmessage "disc too off center": EXIT SUB
    IF ObjectAreaRect.Bottom > SLOimg.Height THEN  Showmessage "disc too off center": EXIT SUB

    IF (ReferenceForceOffsetX <> 0) OR (ReferenceForceOffsetY <> 0) THEN 
        ShowMessage "The disc ROI was shifted (" + _
            STR$(ReferenceForceOffsetX) + ", " + STR$(ReferenceForceOffsetY) + _
            ") to match the reference size"
    END IF


    '-------------------------------------------------------------------
    'set current disc ROI and its subROI dimensions
    '-------------------------------------------------------------------
    DiscROI.Width = ReferenceDiscBMP.Width
    DiscROI.Height = ReferenceDiscBMP.Height
    DiscROI.PixelFormat = pf8bit
    DiscROI.GreyScale
    DiscROI.CopyRect(Dest, SLO_Orig, ObjectAreaRect)
    cv_Bitmap2Mat(DiscROI.Width, DiscROI.Height, 1, DiscROI.Pointer)
    cv_EqualizeHistogram  'cv_EqualizeContrast(CV_NORM_MINMAX)       'CV_NORM_L2


    DEFINT xResolution = 1                                          'divisions have to be 4 byte aligned with 
    DEFINT yResolution = 8                                          'ReferenceImageWidth and ReferenceImageHeight

    SubROI.Width = ReferenceDiscBMP.Width \ xResolution
        IF (SubROI.Width MOD 4) <> 0 THEN SubROI.Width = SubROI.Width + (SubROI.Width MOD 4): d.print "error in X size"
    SubROI.Height = ReferenceDiscBMP.Width \ yResolution
        IF (SubROI.Height MOD 4) <> 0 THEN SubROI.Height = SubROI.Height + (SubROI.Height MOD 4): d.print "error  in Y size"
    SubROI.PixelFormat = pf1bit
    SubROI.GreyScale
    MyImageMatch.template_width = SubROI.Width                   'this holds the sub ROI that searchs
    MyImageMatch.template_height = SubROI.Height                 'shifts within the full ROI
    MyImageMatch.template_bpp = 1
    MyImageMatch.template_lpPixelData = SubROI.Pointer          'make sure this is last

    MyImageMatch.width = ReferenceDiscBMP.Width    'this holds the full optic disc ROI
    MyImageMatch.height = ReferenceDiscBMP.Height
    MyImageMatch.bpp = 1
    MyImageMatch.lpPixelData = ReferenceDiscBMP.Pointer


    DEFINT Successs, x, y
    DEFSTR TheLine

    Dest.Top = 0
    Dest.Left = 0
    Dest.Bottom = SubROI.Height
    Dest.Right = SubROI.Width

    FOR y = 0 TO (DiscROI.Height - SubROI.Height) STEP SubROI.Height \ 4
        FOR x = 0 TO (DiscROI.Width - SubROI.Width) STEP SubROI.Width \ 4
            ObjectAreaRect.Left =   x
            ObjectAreaRect.Top =    y
            ObjectAreaRect.Right  = x + SubROI.Width
            ObjectAreaRect.Bottom = y + SubROI.Height
            SubROI.CopyRect(Dest, DiscROI, ObjectAreaRect)
            MyImageMatch.template_lpPixelData = SubROI.Pointer          'just in case 

            'cross correlation without rotation
            Successs = cv_CrossCorrelateMatch(MyImageMatch)

            IF Successs THEN
                TheLine = STR$(MyImageMatch.centerX) + "," + STR$(MyImageMatch.centerY)
                form.Draw(MyImageMatch.centerX - SubROI.Width\2, MyImageMatch.centerY - SubROI.Height\2, SubROI.BMP)
                '            TheLine = TheLine + STR$(MyImageMatch.rotationDeg) + "," 
                '            TheLine = TheLine + STR$(MyImageMatch.scaleX) + "," + STR$(MyImageMatch.scaleY) + ","
                '            TheLine = TheLine + STR$(MyImageMatch.shear) + "," 
                '            TheLine = TheLine + STR$(MyImageMatch.NumMatchPoints)
                'd.print TheLine
            END IF
        NEXT y
    NEXT x

    Showmessage "showing aligned SubROIdisc"

END SUB





'-------------------------------------------------------------------------------
'  if there is a reference optic disc image, find its match to the 
'  currently loaded optic disc image. The program will then 
'  place a cross where the expected center of the reference disc would match.
'  This allows the user to more accurately align the current scan to the refence. 
'-------------------------------------------------------------------------------


SUB SuggestDiscMatchToReference

    IF (ReferenceDiscBMP.Width <> ReferenceImageWidth) THEN EXIT SUB            'no reference set yet

    DiscROI.Width = SLO_Orig.Width 
    DiscROI.Height = SLO_Orig.Height
    DiscROI.PixelFormat = pf8bit
    DiscROI.GreyScale
    DiscROI.Draw(0, 0, SLO_Orig.BMP)
    cv_Bitmap2Mat(DiscROI.Width, DiscROI.Height, 1, DiscROI.Pointer)
    cv_EqualizeHistogram                                                    'cv_EqualizeContrast(CV_NORM_MINMAX)       'CV_NORM_L2

    'MyImageMatch.detectorType = CV_GRID_ADAPTED_FAST
    'MyImageMatch.detectorParameter = 20                                    'FAST range 5 to 50, higher likely more robust but fewer matches
    'MyImageMatch.detectorParameter2 = 4                                    'grid size
    'MyImageMatch.RANSAC = 10                                               'for findHomography, 0 = off, 1-10 usual values
    '
    'MyImageMatch.detectorType = CV_FAST_DETECTOR
    'MyImageMatch.detectorParameter = 20                                    'FAST typically above 20 otherwise too many matches
    'MyImageMatch.detectorParameter2 = 0
    'MyImageMatch.RANSAC = 10 

    'works relatively good for fundus images...
    MyImageMatch.detectorType = CV_PYRAMID_ADAPTED_FEATUREDETECTOR
    MyImageMatch.detectorParameter = 2                                      'default is 2, pass an integer value (converted by DLL)
    MyImageMatch.detectorParameter2 = 0
    MyImageMatch.RANSAC = 10                                                'for findHomography, 0 = off, 1-10 usual values


    MyImageMatch.MatchMethod = CV_FULL_HOMOGRAPHY
    MyImageMatch.minNumberMatches = 5
    MyImageMatch.RejectPercentage = 0.05
    MyImageMatch.RejectTresholdDist = 3.0
    MyImageMatch.RejectRatioTestValue = 0.8
    MyImageMatch.RejectAngleSD = 0.8
    MyImageMatch.RejectPixelShiftSD = 0.8
    MyImageMatch.RejectPixelDist = 0
    MyImageMatch.PlotResults = False 'True


    MyImageMatch.template_width = ReferenceDiscBMP.Width                    'this holds the ROI that searchs
    MyImageMatch.template_height = ReferenceDiscBMP.Height                  'shifts within the full ROI
    MyImageMatch.template_bpp = 1
    MyImageMatch.template_lpPixelData = ReferenceDiscBMP.Pointer            'make sure this is last

    MyImageMatch.width = DiscROI.Width                                      'this holds the full optic disc ROI
    MyImageMatch.height = DiscROI.Height
    MyImageMatch.bpp = 1
    MyImageMatch.lpPixelData = DiscROI.Pointer

    DEFINT Successs = cv_MatchImages(MyImageMatch)

    DiscMatchCrossClr = clPurple
    IF (Successs = False) THEN
        MyImageMatch.MatchMethod = CV_TM_CCOEFF_NORMED
        Successs = cv_CrossCorrelateMatch(MyImageMatch)
        DiscMatchCrossClr = clBlue
    END IF


    IF Successs THEN
        DrawDiscMatch
        StatusLabel.Caption = "Disc matched by " + IIF(DiscMatchCrossClr = clPurple, "feature detection", "Cross correlation")
    ELSE
        MyImageMatch.centerX = 0
        MyImageMatch.centerY = 0
        StatusLabel.Caption = "Error: cannot match reference disc to image"
    END IF

END SUB



SUB DrawDiscMatch
    Form.Line(MyImageMatch.centerX,      MyImageMatch.centerY - 20,  MyImageMatch.centerX,      MyImageMatch.centerY + 20, DiscMatchCrossClr)
    Form.Line(MyImageMatch.centerX - 20, MyImageMatch.centerY,       MyImageMatch.centerX + 20, MyImageMatch.centerY,      DiscMatchCrossClr)
END SUB

