'===========================================================================
'  Globals and common objects
'===========================================================================

CONST MyVersion AS STRING   = "version 1.4"
CONST TwoPi                 = 2.0! * 3.141592654
CONST Rad2Deg               = 180.0! / 3.141592654
CONST mu AS STRING          = " " + CHR$(181) + "m"
CONST degr AS STRING        = CHR$(176)
CONST CRLF AS STRING        = CHR$(13) + CHR$(10)
CONST MaxN                  = 1000          'maximum number of loci to match
CONST MaxCtrlPts = 150                      'maximum total control points
CONST MaxPts = 10000                        'maximum data points to make up line a single spline segmentation
CONST MaxLayers = 3                         '0 based = 4 layers
CONST cNone                 = -1
CONST cILM                  = 0             'index, internal limiting membrane (ILM) spline and control points
CONST cNFL                  = 1             'index, nerve fiber layer (NFL) spline and control points
CONST cONL                  = 2
'CONST cEZL                  = 3             'index, ellipsoid zone layer
CONST cBM                   = 3             'index, Bruch's membrane
CONST NullResult            = -9999


DEFSTR tmp                                  'general temp string var
DEFINT sPos                                 'starting position for serial search of xml tags
DEFINT cPos                                 'current position while serial searching of xml tags
DEFSTR MyDir                                'keep track of working directory after file save
DEFSTR OutputFileName                       'filename to store analysis output
DEFSTR myFileName                           'dynamic name
DEFSTR OCTFName                             'Filename for OCT image, in case we need to reset modifications
DEFSTR FName                                'first name
DEFSTR LName                                'last name
DEFSTR MRN                                  'Med rec
DEFSTR DOB                                  ' 
DEFSTR ImgType                              'Localizer or OCT of loaded image pairs
DEFSTR Eye                                  'R or L of loaded images
DEFSTR ScanType                             'circle or linear section of loaded images
DEFSTR SLOFocus                             'spherical equivalent of focus
DEFSTR OCTQuality                           'quality score of OCT (1 - 36), like S/N
DEFSTR ARTaverage                           'averaging of scans for final OCT image using automated real-time tracking 
DEFSTR OCTPositionTolerance                 'position tolerance, ART was enabled
DEFINT DeleteKeyDown                        'is the delete key being held down
DEFDBL LastTimeMouseDown                    'patch for double click on QForm
DEFINT LastCtrlPt                           'last active control point
DEFINT OCT_Layer_Picked                     'user clicks to select a segmentation line
DEFINT BusySegmenting                       'prevent build up of slower segmentation routing

DEFINT UnscaledOCTWidth                     'keep original width for conversions
DEFINT UnscaledOCTHeight
DEFINT OCT_Left                             'left side of OCT image for side-by-side plotting
DEFINT SelectedIndx                         'if click and drag on a selected OCT locus
DEFINT DiscHasBeenSet                       'keeping track of disc center set
DEFINT FoveaHasBeenSet                      'useful for keeping old fovea or clicking new location
DEFSNG OCTmmPerXPixels                      'mm per pixel on OCT image in X-dimension 
DEFSNG OCTmmPerYPixels                      'mm per pixel on OCT image in Y-dimension
DEFSNG SLOmmPerXPixels                      'mm per pixel on SLO image in X-dimension 
DEFSNG SLOmmPerYPixels                      'mm per pixel on SLO image in Y-dimension
DEFSNG OCTZoomFactor = 1.50!
DEFINT OpticDiskColor =  RGB(255, 180, 0)
DEFINT FoveaColor =      RGB(0,   228, 255)
DEFINT MatchSegmentClr = RGB(255, 0,   0)

DEFINT SLOflag = False                      'modify SLO image instead of OCT image



TYPE ScanCoodType
    StartX          AS SINGLE               'horiz start of scan from top left of image in mm
    StartY          AS SINGLE               'vert start of scan from top left of image in mm
    EndX            AS SINGLE               'horiz end of scan from top left of image in mm
    EndY            AS SINGLE               'vert end of scan from top left of image in mm
    CenterX         AS SINGLE               'for circle... horiz start of circle from top left of image in mm
    CenterY         AS SINGLE               'for circle... vertical start of circle from top left of image in mm
    Radius          AS SINGLE               'for circle... Radius of scan circle in mm
    Dir             AS INTEGER              'encode either CW or CCW
    Left            AS INTEGER              'integer for rapidq plotting coordinates
    Top             AS INTEGER              'works for both .line and .Circle
    Right           AS INTEGER
    Bottom          AS INTEGER
END TYPE


TYPE PointCoordType
    ScreenX         AS INTEGER
    ScreenY         AS INTEGER
    X               AS SINGLE
    Y               AS SINGLE
END TYPE


TYPE gDataType
    ScreenX         AS INTEGER
    ScreenY         AS INTEGER
    SLOscreenX      AS INTEGER
    SLOscreenY      AS INTEGER
    ScreenThickness AS INTEGER
    X               AS SINGLE
    Y               AS SINGLE
    RNFL_Thickness  AS SINGLE
    ONL_Thickness   AS SINGLE
    Total_Thickness AS SINGLE
    Angle           AS SINGLE
    Num             AS INTEGER
END TYPE



'------------- global objects -----------------------------
DIM OCTLocus(MaxN)      AS gDataType                'get the loci of match points on SLO image
DIM OrigCirc            AS ScanCoodType             'get values at load time before modification
DIM Circ                AS ScanCoodType             'all parameters regarding circular scan
DIM Line                AS ScanCoodType             'all parameters regarding line scan
DIM DiskCenter          AS PointCoordType
DIM FovCenter           AS PointCoordType
DIM WindowLevel         AS PointCoordType
DIM MyOpenDialog        AS QOPENDIALOG
DIM MySaveDialog        AS QSAVEDIALOG
DIM sList               AS QSTRINGLIST
DIM DataList            AS QSTRINGLIST



'------------- global bitmaps -----------------------------

DIM SLO_Orig            AS QBITMAP              'make a backup for window functions
DIM SLOimg              AS QBITMAPEX            'This is full color version showing scan locations, matching loci, fovea, etc
    SLOimg.Width = 1
    SLOimg.Height = 1
    SLOimg.PixelFormat = pf24bit
    'SLOimg.PenSize = 4


DIM OCT_Orig            AS QBITMAP              'keep a backup to cancel user changes
DIM OCTimg              AS QBITMAPEX            'Scaled up, working (user defined) filtered OCT image before segmentation
    OCTimg.Width = 1
    OCTimg.Height = 1
    OCTimg.PixelFormat = pf8bit
    OCTimg.GreyScale


DIM SegmentBMP          AS QBITMAPEX            'copy of OCTimg, but with  more filtering / contrast enhancement for segmentation
    SegmentBMP.Width = 1
    SegmentBMP.Height = 1
    SegmentBMP.Pixelformat = pf8bit
    SegmentBMP.GreyScale


DIM SmallCopy           AS QBITMAPEX            'for segmentation; small copy of original with more filtering / contrast enhancement
DIM EdgeBMP             AS QBITMAPEX            'for segmentation; smaller binary mask for quick search of ILM edges, also used for a retina mask
DIM kMeanCopy           AS QBITMAPEX            'for segmentation; small copy with kMeans segmentation
DIM RetinaMask          AS QBITMAPEX            'for segmentation; aid searching for layers without vitreous or subchoroid
DIM VesselMask          AS QBITMAPEX            'for segmentation; mask containing vertical columns of blood vessel shadows


DIM TweakOCT            AS QBITMAPEX            'full color image with filtered OCT, yellow matching sections, segmentations
    TweakOCT.Width = 1
    TweakOCT.Height = 1
    TweakOCT.PenSize = 2    '4


DIM TempContrastBMP    AS QBITMAPEX            'used for temporary bmp during dynamic contrast / luminance changes
    TempContrastBMP.Width = 1
    TempContrastBMP.Height = 1
    TempContrastBMP.PixelFormat = pf8bit
    TempContrastBMP.GreyScale


DIM LinesBitmap     AS QBITMAPEX                'Full color mask for coloring in  matching sections
    LinesBitmap.PenSize = 4


DIM CircleCursor AS QBITMAP                     'used as a circular cursor for moving control points
    CircleCursor.Width = 56
    CircleCursor.Height = 56
    CircleCursor.FillRect(0, 0, CircleCursor.Width, CircleCursor.Height, 0)
    CircleCursor.Circle(0,0, CircleCursor.Width, CircleCursor.Height, clGreen, 0)
    CircleCursor.Circle(1,1, CircleCursor.Width-1, CircleCursor.Height-1, clGreen, 0)
    CircleCursor.Transparent = True
    CircleCursor.TransparentColor = 0


DIM MyRect  AS QRECT                            'global rect object, use in sub then forget!!!
    MyRect.Top = 0
    MyRect.Left = 0


FUNCTION Strip$(InString AS STRING) AS STRING
    RESULT = LTRIM$(RTRIM$(InString))
END FUNCTION


FUNCTION RoundOff(TheVal AS SINGLE , Accurcy AS INTEGER) AS SINGLE
    IF Accurcy >= 0 THEN
        RoundOff = VAL(STRF$(TheVal , ffFixed , (Accurcy + 2) , Accurcy))
    ELSE
        RoundOff = TheVal
    END IF
END FUNCTION

