
'-------------------------------------------
' for plotting data on a normative template
'-------------------------------------------

$RESOURCE NormBMP AS "Normal_RNFL_Template.bmp"

DEFINT NormXOffset = 68
DEFINT NormYOffset = 431
DEFSNG NormXScale = 1054/360                'convert angle to pixel on graph (pix/micron)
DEFSNG NormYScale = -410/240                'convert microns to pixel on graph (pix/micron)


DIM NormTemplate    AS QBITMAP
    NormTemplate.BMPhandle = NormBMP        'must load it up on init to set sizes
    NormTemplate.Font.Size = 11


CREATE NormForm AS QFORM
    BorderStyle = bsDialog
    ClientWidth = NormTemplate.Width + 8
    ClientHeight = NormTemplate.Height + 8
    OnPaint = NormForm_Paint
    KeyPreview = True
    OnKeyDown = NormForm_KeyDown

    CREATE NormClearButton AS QBUTTON       'QPANEL
        Font.Size = 11
        Caption = "Clear"
        'BevelWidth = 3
        'BevelOuter = bvRaised 
        Width = 75
        Height = 26
        Left = NormForm.ClientWidth - NormClearButton.Width
        Top  = NormForm.ClientHeight - NormClearButton.Height
        OnClick = NormBMPClear
    END CREATE

    CREATE NormRefreshButton AS QBUTTON       'QPANEL
        Font.Size = 11
        Caption = "Refresh"
        'BevelWidth = 3
        'BevelOuter = bvRaised 
        Width = 90
        Height = 26
        Left = NormClearButton.Left - NormRefreshButton.Width - 20
        Top =  NormClearButton.Top
        OnClick = NormForm_Paint
    END CREATE

    CREATE NormSaveButton AS QBUTTON 
        Font.Size = 11
        Caption = "Save Image"
        Width = 130
        Height = 26
        Left = NormRefreshButton.Left - NormSaveButton.Width - 20
        Top =  NormClearButton.Top 
        OnClick = NormForm_Save
    END CREATE

END CREATE



SUB NormForm_KeyDown (Key AS Word, Shift AS INTEGER)
    IF Key = 27 THEN NormForm.Close    'ESC
END SUB



SUB NormCalculate
    ' line sections are calcuated real-time, spline segmentation may not yet be done
    ' so convert segmentation lines to RNFL thickness, this will clear all 'lines' segment measurements!!
    IF ScanType = "CIRCLE" AND (SplineButton.Caption = "Spline") THEN ParseSegmentation
END SUB



SUB ToggleNormTemplate
    IF NormForm.Visible THEN 
        NormForm.Close 
    ELSE
        NormCalculate
        NormForm.Show
    END IF
END SUB



SUB NormBMPClear
    NormTemplate.BMPhandle = NormBMP            'reload template
    NormForm.Draw(0, 0, NormTemplate.BMP)
END SUB



SUB NormForm_Paint
    NormCalculate                                       'reload new measurements if control points are moved
    DEFINT i, x, y, overfl
    DEFSNG myAng
    FOR i = 0 TO OCTLocus(0).Num -1                     '0 based arrays
        IF OCTLocus(i).RNFL_Thickness > 0 THEN
            myAng = OCTLocus(i).Angle
            IF myAng < 0 THEN myAng = myAng + 360
            IF myAng > 360 THEN myAng = myAng - 360
            x = NormXOffset + (NormXScale * myAng)
            y = NormYOffset + (NormYScale * OCTLocus(i).RNFL_Thickness)
            IF y < 1 THEN
                y = 1
                overfl = True
            ELSE
                overfl = False
            END IF
            NormTemplate.Circle(x - 3, y - 3, x + 3, y + 3, 0, 0)
            IF overfl THEN NormTemplate.TextOut(x, y + 10, "(" + STRF$(OCTLocus(i).RNFL_Thickness, ffFixed, 10, 1) + ")", 0, clWhite)
        END IF
    NEXT i
    NormForm.Draw(0, 0, NormTemplate.BMP)
END SUB


SUB NormForm_Save
    MySaveDialog.Filter = "bitmap file|*.bmp"
    MySaveDialog.FilterIndex = 0
    MySaveDialog.FileName = LName + "_" + FName + "_RNFL_" + Eye + ".bmp"
    IF MySaveDialog.Execute THEN 
        NormForm_Paint                                      'refresh measurements and bitmap
        NormTemplate.SaveToFile(MySaveDialog.FileName)
    END IF
END SUB