
'--------------------------------------------------------------------------------------------
'  if user sets one scan as a reference, they can use 
'  the reference to compare all subsequent scans in a session
'--------------------------------------------------------------------------------------------


DEFSTR TheReferenceFolder = Application.Path + "\References\"
CONST ReferenceImageWidth = 256                         'must be 4 byte aligned!!!
CONST ReferenceImageHeight = 256                        'must be 4 byte aligned!!!
DIM ReferenceCanBeSaved         AS INTEGER              'set yet?
DIM ReferenceName               AS STRING
DIM ReferenceForceOffsetX       AS INTEGER
DIM ReferenceForceOffsetY       AS INTEGER
DIM ReferenceDiskScreenX        AS SINGLE
DIM ReferenceDiskScreenY        AS SINGLE
DIM ReferenceDiskX              AS SINGLE
DIM ReferenceDiskY              AS SINGLE
DIM ReferenceFovX               AS SINGLE
DIM ReferenceFovY               AS SINGLE
DIM ReferenceFovScreenX         AS INTEGER
DIM ReferenceFovScreenY         AS INTEGER
'DIM ReferenceDisk               AS PointCoordType      'Struct causes a memory leak!!!
'DIM ReferenceFov                AS PointCoordType      'Struct causes a memory leak!!!
DIM ReferenceDiscBMP            AS QBITMAPEX            'get full optic disc area needed
ClearReferenceData                                      'need to persist for entire session
'--------------------------------------------------------------------------------------------


CREATE ReferencePanel AS QPANEL
    Parent = Form
    Left = 790
    Top = 865
    Width = 600
    Height = 48
    Color = RGB(135, 104, 21)

    CREATE SetAsReferenceButton AS QBUTTON
        Left = 40 
        Top = 8
        Width = 140
        Height = 34
        Hint = "Set this as reference for scans in this session"
        ShowHint = True
        Caption = "Set As Reference"
        OnClick = SetCurrentAsReference
    END CREATE

    CREATE Load_RemoveReferenceButton AS QBUTTON
        Left = 200
        Top = 8
        Width = 140
        Height = 34
        Tag = False         'no reference loaded
        Hint = "Load a reference image for matching scan center"
        ShowHint = True
        Caption = "Load reference"
        OnClick = LoadRemoveReference
    END CREATE

    CREATE ViewReferenceButton AS QBUTTON
        Left = 360
        Top = 8
        Width = 140
        Height = 34
        Hint = "view current reference image, if set"
        ShowHint = True
        Caption = "View reference"
        OnClick = ViewReference
        Enabled = False
    END CREATE

END CREATE              'ReferencePanel




CREATE RefPreviewForm AS QFORM
    Caption = "Reference"
    Font.Size = 11
    Top = -20               'flag, set to negative if user has not positioned form
    Left = -20
    ClientWidth = ReferenceImageWidth 
    ClientHeight = ReferenceImageHeight
    OnPaint = RefPreview_Paint
END CREATE





SUB ClearReferenceData
    ReferenceDiscBMP.Width = 4                                      'flag not set yet
    ReferenceDiscBMP.Height = 4     
    ReferenceCanBeSaved = False
    ReferenceDiskX          = 0.0
    ReferenceDiskY          = 0.0
    ReferenceDiskScreenX    = 0
    ReferenceDiskScreenY    = 0
    ReferenceFovX           = 0.0
    ReferenceFovY           = 0.0
    ReferenceFovScreenX     = 0
    ReferenceFovScreenY     = 0
END SUB






SUB LoadRemoveReference

    IF Load_RemoveReferenceButton.Tag THEN                                          'flag reference is loaded

        ClearReferenceData
        Load_RemoveReferenceButton.Tag = False                                      'flag, reference is cleared
        Load_RemoveReferenceButton.Caption = "Load reference"
        Load_RemoveReferenceButton.Hint = "Load a reference image for matching scan center"
        ViewReferenceButton.Enabled = False

    ELSE

        ReferenceName = GenerateReferenceName
        DEFSTR TheFileName = TheReferenceFolder + ReferenceName + ".csv"

        IF FILEEXISTS(TheFileName) THEN
            DataList.LoadFromFile( TheFileName )
        ELSE
            MyOpenDialog.FileName = GenerateReferenceName 
            MyOpenDialog.Filter = "Input Data|*.csv|Images|*.bmp|All Files|*.*"
            MyOpenDialog.FilterIndex = 0
            IF MyOpenDialog.Execute THEN
                DataList.LoadFromFile( MyOpenDialog.FileName )
                ReferenceName = FILENAMENOEXT( STRIPFILENAME (MyOpenDialog.FileName))   'set this!!
            ELSE
                EXIT SUB
            END IF
        END IF

        '------ minimial error checking !!  -----------------
        IF DataList.ItemCount < 3 THEN 
            ShowMessage "Error loading reference"
            EXIT SUB
        END IF

        DEFINT i
        DEFSNG TheRadius
        FOR i = 1 TO DataList.ItemCount-1
            IF TALLY(DataList.Item(i), ",") > 1 THEN
                ReferenceFovX = VAL(FIELD$(DataList.Item(i), ",", 7))           'field "RNFL_Angle"
                ReferenceFovY = VAL(FIELD$(DataList.Item(i), ",", 8))           'field "RNFL_Thickness_um"
                ReferenceFovScreenX  = ReferenceFovX / SLOmmPerXPixels
                ReferenceFovScreenY  = ReferenceFovY / SLOmmPerYPixels

                ReferenceDiskX = VAL(FIELD$(DataList.Item(i), ",", 11))         'field "DiscX_mm" on reference SLO bmp
                ReferenceDiskY = VAL(FIELD$(DataList.Item(i), ",", 12))         'field "DiscY_mm" on reference SLO bmp
                ReferenceDiskScreenX = VAL(FIELD$(DataList.Item(i), ",", 13))   'field "Xpix_on_SLO"
                ReferenceDiskScreenY = VAL(FIELD$(DataList.Item(i), ",", 14))   'field "Ypix_on_SLO"
                TheRadius = VAL(FIELD$(DataList.Item(i), ",", 6)) / 1000.0!     'uses "NewScanRadius_um" in case user modified
                EXIT FOR
            END IF
        NEXT i

        IF Circ.Radius < 400 THEN                                               'user has not loaded a circular scan!
            IF MessageBox("radius of circle scans not set. Use reference value of " +  STRF$(TheRadius, ffGeneral, 4, 4) + " mm ?", "Set radius?", &H30 + 4) = mrNo THEN
                EXIT SUB
            ELSE
                Circ.Radius = TheRadius
            END IF
        END IF

        IF ( ((TheRadius / Circ.Radius) > 1.1) OR  ((TheRadius / Circ.Radius) < 0.9) ) THEN
            IF MessageBox("radii of circle scans do not match. Continue?", "Cancel Reference?", &H30 + 4) = mrNo THEN
                ClearReferenceData
                EXIT SUB
            END IF
        END IF

        IF PaintReferenceOnNorms(DataList) = False THEN
            ShowMessage "Unknown error plotting reference"
        END IF

        TheFileName = TheReferenceFolder + ReferenceName + ".bmp"
        IF FILEEXISTS(TheFileName) THEN
            ReferenceDiscBMP.LoadFromFile(TheFileName)
            DEFINT w = ReferenceDiscBMP.Width
            DEFINT h = ReferenceDiscBMP.Height
            IF (w MOD 4) <> 0 THEN ReferenceDiscBMP.Width = w - (w MOD 4)           'must be 32 bit aligned
            IF (h MOD 4) <> 0 THEN ReferenceDiscBMP.Height = h - (h MOD 4)          'must be 32 bit aligned
            Load_RemoveReferenceButton.Caption = "Remove reference"
            Load_RemoveReferenceButton.Hint = "Clear current reference image"
            ViewReferenceButton.Enabled = True
            Load_RemoveReferenceButton.Tag = True                                   'flag, reference is loaded
            SetFoveaLocus(ReferenceFovScreenX, ReferenceFovScreenY)               'update officially
            'now FovCenter.X = ReferenceFovX
            'now FovCenter.Y = ReferenceFovY
            SuggestDiscMatchToReference                                             'find best match to optic disc
            ShowMessage "Loaded reference data and image for " + ReferenceName
            StatusLabel.Caption = "Reference Scan radius = " + STRF$( Circ.Radius * 1000, ffGeneral, 5, 6) + mu
            FormRepaint
        END IF

    END IF

END SUB




SUB SetCurrentAsReference

    IF ScanType <> "CIRCLE" THEN 
        Showmessage "Only circular scans are used as a reference"
        EXIT SUB
    END IF

    IF (FoveaHasBeenSet = False) OR (FovCenter.ScreenX = NullResult) THEN
        Showmessage "manually set the fovea location first"
        EXIT SUB
    END IF


    'load unmodified parameters from original load time
    ReferenceName = GenerateReferenceName
    
    MEMCPY (UDTPTR(Circ), UDTPTR(OrigCirc), SIZEOF(OrigCirc))                       'copy over the orignal values for current analysis
    ScanRadiusEdit.Text = ""                                                        'forces update of radius to current scan
    UpdateScanRadius


    'need to use the center of the circular scan from load time
    ReferenceDiskX = OrigCirc.CenterX
    ReferenceDiskY = OrigCirc.CenterY
    ReferenceDiskScreenX = ReferenceDiskX / SLOmmPerXPixels
    ReferenceDiskScreenY = ReferenceDiskY / SLOmmPerYPixels
    SetOpticDiskCenter (ReferenceDiskScreenX, ReferenceDiskScreenY)                'update officially
    ReferenceDiskX = DiskCenter.X                                                  'update after setting
    ReferenceDiskY = DiskCenter.Y                                                  'update after setting
    DiscHasBeenSet = True


    'after disc center has been set, fovea angle must be calculated after seting disc center
    ReferenceCanBeSaved = True
    ReferenceFovScreenX  = FovCenter.ScreenX 
    ReferenceFovScreenY  = FovCenter.ScreenY
    SetFoveaLocus(ReferenceFovScreenX, ReferenceFovScreenY)                        'update officially
    ReferenceFovX  = FovCenter.X                                                   'update after setting
    ReferenceFovY  = FovCenter.Y 

    'these values are critical to the match - search 
    ReferenceDiscBMP.Width = ReferenceImageWidth                                    'must be 4 byte aligned!!!
    ReferenceDiscBMP.Height = ReferenceImageHeight                                  'must be 4 byte aligned!!!
    ReferenceDiscBMP.PixelFormat = pf8bit
    ReferenceDiscBMP.GreyScale

    DEFINT x = ReferenceDiskScreenX - ReferenceDiscBMP.Width\2
    DEFINT y = ReferenceDiskScreenY - ReferenceDiscBMP.Height\2

    DEFINT InBounds = True                                                          'Check boundary
    IF x < 0 THEN x = 0
    IF y < 0 THEN y = 0
    IF (x + ReferenceDiscBMP.Width) > SLO_Orig.Width THEN
        x = SLO_Orig.Width - ReferenceDiscBMP.Width
        InBounds = False
    END IF
    IF (y + ReferenceDiscBMP.Height) > SLO_Orig.Height THEN
        x = SLO_Orig.Height - ReferenceDiscBMP.Height
        InBounds = False
    END IF

    'store the reference image
    IF (InBounds = False) THEN
        IF MessageBox("Reference image error: image out of bounds", "Continue ?", &H30 + 4) = mrNo THEN
            ClearReferenceData
            EXIT SUB
        END IF
    END IF

    ReferenceDiscBMP.Draw(-x, -y, SLO_Orig.BMP)
    cv_Bitmap2Mat(ReferenceDiscBMP.Width, ReferenceDiscBMP.Height, 1, ReferenceDiscBMP.Pointer)
    cv_EqualizeHistogram                'cv_EqualizeContrast(CV_NORM_MINMAX)       'CV_NORM_L2


    GetMatchingLoci
    FormRepaint
    ViewReferenceButton.Enabled = True

END SUB





SUB RefPreview_Paint
    RefPreviewForm.Draw(0, 0, ReferenceDiscBMP.BMP)
    RefPreviewForm.FillRect((ReferenceDiscBMP.Width \ 2) - 4, (ReferenceDiscBMP.Height \ 2) - 4,_
                            (ReferenceDiscBMP.Width \ 2) + 4, (ReferenceDiscBMP.Height \ 2) + 4, clYellow)
    IF ReferenceDiscBMP.Width = ReferenceImageWidth THEN
        RefPreviewForm.TextOut(1, 1, "Reference image", clRed, clBlack)
    ELSE
        RefPreviewForm.TextOut(8, 8, "No Reference set", clRed, clBlack)
    END IF
END SUB





SUB ViewReference
    IF RefPreviewForm.Visible THEN
        RefPreviewForm.Close
    ELSE
        RefPreviewForm.Top = ReferenceDiskScreenY - ReferenceDiscBMP.Height \ 2
        RefPreviewForm.Left = OCT_Left                                  '10 + ReferenceDiskScreenX + ReferenceDiscBMP.Width \ 2 
        IF RefPreviewForm.Top < 0 THEN RefPreviewForm.Top = 20          'user has not set form position
        RefPreviewForm.Show
    END IF

END SUB






SUB StoreReferenceData

    DEFSTR c = ","
    DEFSTR sTmp = ""

    MySaveDialog.Caption = "Save Reference Data?"
    MySaveDialog.InitialDir = TheReferenceFolder
    MySaveDialog.Filter = "Output files|*.txt;*.csv|All Files|*.*"
    MySaveDialog.FilterIndex = 0
    MySaveDialog.FileName = ReferenceName
    IF MySaveDialog.Execute THEN IF MySaveDialog.FileName = "" THEN EXIT SUB                 'user bails out
    sTmp = MySaveDialog.FileName
    ReferenceName = FileNameNoExt(sTmp)

    DataList.Clear
   'create a header
        sTmp = _
        "LastName,FirstName," + _
        "MRN," + _
        "Eye," + _
        "OriginalRadius_um,Ref_NewScanRadius_um," + _
        "Ref_RNFL_Angle" + degr + "," + _
        "Ref_RNFL_Thickness_um," + _
        "Ref_ONL_Thickness_um," + _
        "Ref_Total_Thickness_um," + _
        "Ref_DiscX_mm,Ref_DiscY_mm," + _
        "Ref_Xpix_on_SLO,Ref_Ypix_on_SLO," + _
        "Ref_Fovea_X, Ref_Fovea_Y," + _
        "Ref_Fovea_ang" + degr + "," + _
        "Ref_Focus, Ref_Quality,Ref_ARTaverage," + _
         "Ref_Met_PositionTolerance?" + CRLF
    DataList.Text = sTmp

    ParseSegmentation           'convert segmentation lines to RNFL thickness, this will clear all 'lines' segment measurements!!

    DEFINT i
    FOR i = 0 TO OCTLocus(0).Num -1                                                 '0 based arrays
        IF OCTLocus(i).RNFL_Thickness > 0 THEN
            sTmp = sTmp + LName + c + FName + c + MRN + c + Eye + c                 'name, MRN, and eye
            IF ScanType = "CIRCLE" THEN                                             'origninal scan radius
                sTmp = sTmp + STRF$(Circ.Radius * 1000.0!, ffFixed, 10, 1)  + c
            ELSE
                sTmp = sTmp + "Section" + c
            END IF
            sTmp = sTmp + ScanRadiusEdit.Text + c                                   'new scan radius
            sTmp = sTmp + STRF$( OCTLocus(i).Angle, ffFixed, 10, 3) + c             'angle of thickness measurment
            sTmp = sTmp + STRF$( OCTLocus(i).RNFL_Thickness, ffFixed, 10, 1) + c    'RNFL thickness at that angle
            sTmp = sTmp + STRF$( OCTLocus(i).ONL_Thickness, ffFixed, 10, 1) + c     'ILM to ONL thickness at that angle
            sTmp = sTmp + STRF$( OCTLocus(i).Total_Thickness, ffFixed, 10, 1) + c   'ILM to BM thickness at that angle
            sTmp = sTmp + STRF$( ReferenceDiskX, ffFixed, 10, 1) + c               'location of disk (mm) on SLO
            sTmp = sTmp + STRF$( ReferenceDiskY, ffFixed, 10, 1) + c               'in X, Y coordinates 0,0 at top left
            sTmp = sTmp + STRF$( OCTLocus(i).SLOscreenX , ffFixed, 10, 1) + c       'location of measurement on SLO in pixels
            sTmp = sTmp + STRF$( OCTLocus(i).SLOScreenY, ffFixed, 10, 1) + c        'in X, Y coordinates 0,0 at top left
            sTmp = sTmp + STRF$( ReferenceFovX, ffFixed, 10, 1) + c                'location of fovea on SLO in pixels
            sTmp = sTmp + STRF$( ReferenceFovY, ffFixed, 10, 1) + c                'in X, Y coordinates 0,0 at top left
            sTmp = sTmp + STRF$(FoveaAngle(), ffFixed, 10, 1) + c                   'angle of fovea from disk center (inferior to disk = negative angle)
            sTmp = sTmp + SLOFocus + c                                              'focus of SLO
            sTmp = sTmp + OCTQuality + c                                            'OCT image quality score 
            sTmp = sTmp + ARTaverage + c                                            'average number of images making up OCT with automated real-time tracking
            sTmp = sTmp + OCTPositionTolerance + CRLF                               'OCT acquired within position tolerance?
        END IF
    NEXT i
    sTmp = sTmp + CRLF

    DataList.Text = sTmp
    DataList.SaveToFile(ReferenceName + ".csv")
    ReferenceDiscBMP.SaveToFile(ReferenceName + ".bmp")
    ReferenceCanBeSaved = False
END SUB




FUNCTION PaintReferenceOnNorms(inList AS QSTRINGLIST) AS INTEGER
    RESULT = False

    DEFINT i, x, y, overfl
    DEFSNG myAng, MyRNFLThickness

    FOR i = 1 TO inList.ItemCount-1
        IF TALLY(inList.Item(i), ",") > 1 THEN
            myAng =             VAL(FIELD$(inList.Item(i), ",", 7))
            MyRNFLThickness =   VAL(FIELD$(inList.Item(i), ",", 8))
            IF MyRNFLThickness > 0 THEN
                IF myAng < 0 THEN myAng = 360 + myAng
                x = NormXOffset + NormXScale * myAng
                y = NormYOffset + NormYScale * MyRNFLThickness

                IF y < 1 THEN
                    y = 1
                    overfl = True
                ELSE
                    overfl = False
                END IF
                NormTemplate.Circle(x - 3, y - 3, x + 3, y + 3, RGB(87, 87, 87), RGB(167, 167, 167))
                IF overfl THEN NormTemplate.TextOut(x, y + 10, "(" + STRF$(MyRNFLThickness, ffFixed, 10, 1) + ")", 0, clWhite)
            END IF
        END IF
    NEXT i

    NormForm.Draw(0, 0, NormTemplate.BMP)
    RESULT = True

END FUNCTION




FUNCTION GenerateReferenceName() AS STRING
    STATIC Reslt AS STRING
    Reslt = FName + "_" + LName + "_" + Eye + "_" + MRN + "_" + "Reference"      'keep as static
    RESULT = Reslt
END FUNCTION
