
'------------------------------------------------------------
' Segmentation routines
'------------------------------------------------------------

DEFSNG Scale_Factor = 4.0!                          'use shrink factor for speed ...


'------------------------------------------------------------
'   Arrays for fitting layers with polynomial curve
'------------------------------------------------------------
CONST MaxPolyArryNum AS INTEGER = MaxCtrlPts        'should be the same as # of control points
CONST MaxPolyOrder AS INTEGER = 6                   'add one more than you think is necessary
DEFDBL InFitDataX(MaxPolyArryNum) 
DEFDBL InFitDataY(MaxPolyArryNum) 
DEFDBL OutputCoeff(MaxPolyOrder)



'------------------------------------------------------------
'   coordinates for summing to find edges lines
'------------------------------------------------------------

DEFINT NumVertFilterElements = 2
DEFINT VertFilterX(NumVertFilterElements) = {_
0, 0, 0}
DEFINT VertFilterY(NumVertFilterElements) = {_
-1, 0, 1}


DEFINT NumEdgeFilterElements = 23
DEFINT EdgeFilterX(NumEdgeFilterElements) = {_
-3, -2, -1, -1, 0, 0, 1, 1 , 2, 3,_
-5, -4, -3, -2, -1, -1, 0, 0, 1, 1, 2, 3, 4, 5}

DEFINT EdgeFilterY(NumEdgeFilterElements) = {_
1, 1, 1, 1, 1, 1, 1, 1, 1, 1,_
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}


'DEFINT NumEdgeFilterElements = 21
'DEFINT EdgeFilterX(NumEdgeFilterElements) = {_
'-2, -1, 0, 1, 2,_
'-3, -2, -1, 0, 1, 2, 3,_
'-4, -3, -2, -1, 0, 1, 2, 3, 4}
'
'DEFINT EdgeFilterY(NumEdgeFilterElements) = {_
'2, 2, 2, 2, 2,_
'1, 1, 1, 1, 1, 1, 1,_
'0, 0, 0, 0, 0, 0, 0, 0, 0}




'------------------------------------------------------------------------------
'   setup dimensions for smaller pre-filtered original 
'   all filtering and searching works much more efficiently on smaller BMP
'   also -- cv_Resize(wid as integer, hei as integer, interpFlag as integer)
'------------------------------------------------------------------------------

SUB PrepareSmallImages(inBMP AS QBITMAP)

    myRect.Left = 0
    myRect.Top = 0
    myRect.Right =  (inBMP.Width \ Scale_Factor)  - ((inBMP.Width \ Scale_Factor) MOD 4)
    myRect.Bottom = (inBMP.Height \ Scale_Factor) - ((inBMP.Height \ Scale_Factor) MOD 4)

    kMeanCopy.Width = myRect.Right
    kMeanCopy.Height = myRect.Bottom
    kMeanCopy.PixelFormat = pf8bit
    kMeanCopy.GreyScale

    EdgeBMP.Width = myRect.Right
    EdgeBMP.Height = myRect.Bottom
    EdgeBMP.PixelFormat = pf8bit
    EdgeBMP.GreyScale

    RetinaMask.Width = myRect.Right
    RetinaMask.Height = myRect.Bottom
    RetinaMask.PixelFormat = pf8bit
    RetinaMask.GreyScale

    VesselMask.Width = myRect.Right
    VesselMask.Height = myRect.Bottom
    VesselMask.PixelFormat = pf8bit

    SmallCopy.Width = myRect.Right
    SmallCopy.Height = myRect.Bottom
    SmallCopy.PixelFormat = pf8bit
    SmallCopy.GreyScale
    SmallCopy.StretchDraw(myRect, inBMP.BMP)
END SUB



'*** make sure we have white RNFL on black vitreous ***
SUB CheckReverseContrast
    DEFINT x = 0,  y = 0
    DO                                                          'do-LOOP faster under RQ
        INC(y, &HFF AND SegmentBMP.Pixel(x, 2))                 'sample upper left 32 pixels
        INC(y, &HFF AND SegmentBMP.Pixel(x, 3))
        x++
    LOOP UNTIL x > 16

    IF (y\32) > &H80 THEN                                       '128 intensity cutoff
        SegmentBMP.CopyMode = cmPatInvert                       'Combines  source with the image using XOR operator 
        SegmentBMP.Draw(0, 0, SegmentBMP.BMP)                   'invert here
        SegmentBMP.CopyMode = cmSrcCopy                         '******* now BMP is white RNFL on black vitreous ***
    END IF
END SUB




SUB FilterOCTForSegmentation

    SegmentBMP.Width = OCTimg.Width
    SegmentBMP.Height = OCTimg.Height
    SegmentBMP.PixelFormat = pf8bit
    SegmentBMP.GreyScale
    SegmentBMP.Draw(0, 0, OCTimg.BMP)                           'copy original OCT BMP

    CheckReverseContrast

    '----------------------------------------------------------------------------------
    'median filtering and enhance contrast to remove background noise before re-scaling
    'low pass filtering, a critical step for layer segmentation due to noise
    '      Changes are permanent to SegmentBMP pixel data!!!
    '----------------------------------------------------------------------------------
    cv_Bitmap2Mat(SegmentBMP.Width, SegmentBMP.Height, 1, SegmentBMP.Pointer)       'first store it in OpenCV interface


    DEFINT Krnl = 5                                                                 'don't take this much higher
    cv_Blur(Krnl , Krnl)
    cv_MedianBlur(Krnl + 2)


    '------------------------------------------------------------------------------------------
    'contrast enhancement, a critical step to remove background speckle
    ' cv_EqualizeHistogram  <-contrast gain but enhances vitreous particles/artifact too much
    '------------------------------------------------------------------------------------------
    DEFDBL alpha                                    'need contrast value in double var for pointer
    DEFDBL beta                                     'need brightness value in double var for pointer
    alpha = 2.0
    beta = -60      '-30 ok

    IF AdaptiveThresholdCheckBox.Checked THEN       'Adaptive filtering removes intra layer noise, but needs more contrast to remove detection of vitreous particles
        alpha = 1.8                                 'contrast value, default = 1.0     higher values remove low reflectance tissue
        beta = -80                                  'brightness value for dropping speckle and vitreous noise
    END IF


'    IF ImageExponentCheckBox.Checked THEN 
'        DEFDBL ThePower = VAL(ImageExponentEdit.Text)
'        IF (ThePower > 1.0) AND (ThePower < 5.0) THEN
'            cv_Pow(ThePower)
'        ELSE
'            ImageExponentCheckBox.Checked = False
'            ImageExponentEdit.Text = 2
'        END IF
'    END IF

'    IF ImageExponentCheckBox.Checked THEN
'            alpha = 2.0'1.5 
'            beta = RNFL_ThresholdTrackBar.Position - 50: d.print RNFL_ThresholdTrackBar.Position- 50 '0.0
'    END IF
'
    cv_ContrastWindow(alpha , beta)

    '------------------------------------------------------------------------------
    'Setup smaller BMP for ILM Canny filter etc, to be efficient
    ' Copy of full size is stored in SmallCopy.BMP
    '------------------------------------------------------------------------------
    PrepareSmallImages(SegmentBMP)

    '------------------------------------------------------------------------------
    'filtering for SmallCopy BMP used for segmenting layers without speckle noise
    'median blur absolutely needed, 3,5,7 are good working values
    '------------------------------------------------------------------------------
    cv_Bitmap2Mat(SmallCopy.Width , SmallCopy.Height , 1 , SmallCopy.Pointer)
    cv_MedianBlur(3)

    '------------------------------------------------------------------------------
    ' kMean segmentation to cluster intensity for each layer
    '------------------------------------------------------------------------------
    kMeanCopy.Draw(0, 0, SmallCopy.BMP)
    cv_Bitmap2Mat(kMeanCopy.Width , kMeanCopy.Height , 1 , kMeanCopy.Pointer)
    cv_kMeans(5, 10)           '1st is num of intensity levels, 2nd is max iterations to randomize

    '------------------------------------------------------------------------------
    '  thresholding on EdgeBMP to quick find edges for ILM, BM
    '  uses either OTSU or adaptive thresholding
    '  works as a mask to for isolating retina from  vitreous space / sub choroid
    '------------------------------------------------------------------------------
'    IF UseKmeansCheckBox.Checked THEN
'        EdgeBMP.Draw(0, 0, kMeanCopy.BMP)
'    ELSE
'        EdgeBMP.Draw(0, 0, SmallCopy.BMP)
'    END IF


    EdgeBMP.Draw(0, 0, kMeanCopy.BMP)                           'use kMeans for detecting edges
    cv_Bitmap2Mat(EdgeBMP.Width, EdgeBMP.Height, 1, EdgeBMP.Pointer)
    IF AdaptiveThresholdCheckBox.Checked THEN                   'Adaptive filtering can remove intra layer noise, but also enhances detection of vitreous particles ...
        cv_Blur(11, 11)                                         'VAL(LowPassKernalEdit.Text), VAL(LowPassKernalEdit.Text))
        Krnl = 21                                               '12-25  works well '=  ThresholdTrackBar.Position + 20; IF ((Krnl MOD 2) = 0) THEN INC(Krnl, 1)                     'keep size odd
        cv_adaptiveThreshold(255, CV_ADAPTIVE_THRESH_GAUSSIAN_C, CV_THRESH_BINARY, Krnl, 0)
        cv_Erode(CV_MORPH_ELLIPSE, 7)                           'helps smooth and remove isolated dots
        cv_CannyThreshold(200, 20, 3)                           'edge detection for ILM and BM, some of RNFL, and choroid as well
        cv_Dilate(CV_MORPH_ELLIPSE, 7)                          'fill in gaps
    ELSE
        cv_Erode(CV_MORPH_ELLIPSE, 3)                           'helps smooth and remove isolated dots
        cv_OtsuThreshold (CV_THRESH_BINARY)                     'good thresholding with speckle noise
        cv_Dilate(CV_MORPH_ELLIPSE, 5)                          'fill in gaps
    END IF


    '------------------------------------------------------------------------------
    '  RetinaMask, mixture of mask using EdgeBMP and OCT copy
    '  Can use this to aid searching for layers without vitreous or subchoroid
    '------------------------------------------------------------------------------

    RetinaMask.CopyMode = cmsrccopy
    RetinaMask.Draw(0, 0, SmallCopy.BMP)                        'probably should be free of kmeans segmentation
    RetinaMask.CopyMode = cmSrcAnd
    RetinaMask.Draw(0, 0, EdgeBMP.BMP)

    GetVesselMask
END SUB



'------------------------------------------------------------------------------
'  VesselMask, create a mask of vertical lines aligned with vert shadows 
' use this to avoid segmenting an area with a vessel
'------------------------------------------------------------------------------

SUB GetVesselMask

    IF UseVesselMaskCheckBox.Checked THEN 
        '    DEFDBL Biass
        '    'default VesselThreshTrackbar.Position = 40
        '    IF VesselSearchWindowSize = 1 THEN Biass = 0
        '    IF VesselSearchWindowSize = 2 THEN Biass = 10
        '    IF VesselSearchWindowSize > 2 THEN Biass = 20
        '    DEFDBL vesselThresh = (Biass + (VesselThreshTrackbar.Position) * 2.00) / 1000.00     'd.print VesselThreshTrackbar.Position * 2          ' 80 - 110 works well
        DEFINT VesselSearchWindowSize = 1
        DEFDBL vesselThresh = 0.065
        cv_FindVesselShadows(SmallCopy.Width, SmallCopy.Height, 1, SmallCopy.Pointer, VesselMask.Pointer, 30, VesselSearchWindowSize, vesselThresh)
        '   cv_FindVesselShadows(kMeanCopy.Width, kMeanCopy.Height, 1, kMeanCopy.Pointer, VesselMask.Pointer, 30, VesselSearchWindowSize, vesselThresh)
    ELSE
        VesselMask.FillRect(0, 0, VesselMask.Width, VesselMask.Height, 0)
    END IF
END SUB


'---------------------------------------------------------
'Main algorithm to detect retinal layers using OpenCV

'segmentation algorithm:
'1- bitmaps converted to 8-bit greyscale, copy of original OCT
'2- median filter, added on top of users preference
'3- increase contrast, decrease brightness, reduces noise
'4- copy and resize OCT image by Scale_Factor to speed processing
'5- median filter again of smaller OCT
'6- adaptive filter (or OTSU thresholding) of smaller OCT
'7- Canny edge filter of smaller OCT
'8- RNFL, ILM ellipsoid zone, Bruch's membrane detection uses binary vertial search on smaller bmp
'9- using the positions from smaller OCT fine tune detection of intensity change in full size OCT copy
' adding or replacing with gamma (0.8 to 1.1) and/or heavy blur typically show no improvment
'---------------------------------------------------------


SUB SegmentOCTimage

    IF BusySegmenting THEN EXIT SUB                             'avoid multitasking event pile up
    BusySegmenting = True
    DEFSTR ErrMessgage = ""

    LockAllSplines(True)                                        'lock and prevent user drawn manual segmentation
    FilterOCTForSegmentation

    'first get ILM, most robust and serves as guide for searching RNFL
    IF SegmentILM < 1 THEN              'most further segmentation dependent upon this one
        BusySegmenting = False
        ErrMessgage = ErrMessgage + "no detectable ILM signal" + CRLF
    END IF

    IF SegmentRNFL < 1 THEN
        BusySegmenting = False
        ErrMessgage = ErrMessgage + "Error in RNFL segmentation" + CRLF
    END IF

    IF SegmentONL < 1 THEN              'This depends on RNFL segmentation
        BusySegmenting = False
        ErrMessgage = ErrMessgage + "Error in ONL segmentation" + CRLF
    END IF

'    IF SegmentEZL < 1 THEN
'        BusySegmenting = False
'        ErrMessgage = ErrMessgage + "Error in Ellipsoid Zone segmentation" + CRLF
'    END IF

    IF SegmentBML < 1 THEN
        BusySegmenting = False
        ErrMessgage = ErrMessgage + "Error in Bruch's membrane segmentation" + CRLF
    END IF

    IF ErrMessgage <> "" THEN 
        ShowMessage ErrMessgage
        EXIT SUB
    END IF

    ShowSplineControlPoints(False)              'let the user select
    CalcAllLayerSplines
    BusySegmenting = False
    DrawOCTSegmentation
END SUB







'=============================================
' do ILM segmentation on global EdgeBMP
' just about all layers are dependent
' upon successful segmentation of ILM !!!
'=============================================

FUNCTION SegmentILM() AS INTEGER
    DEFINT EdgeCropSize = 6                                 'set crop of top and bottom edges for EdgeBMP
    DEFINT StrtY = EdgeCropSize                             'start top vertical offset
    DEFINT EndY = EdgeBMP.Height - EdgeCropSize             'end of vertical search
    DEFINT i , j                                            'loopers
    DEFINT xd, yd                                           'temp vars for looping and rescaling
    DEFINT x, y = StrtY                                     'coordinates on edge filtered image
    DEFINT yLmit                                            'limit of searchs in vertical direction
    DEFINT MaxILMChange = 40                                'maximum change in pixels that the ILM can change in horizontal dimension
    DEFINT SearchRes = 4                                    '+/- size in absolute pixels to fine tune search for edges
    CurrLayer = cILM                                        'setting  CurrLayer = ILM
    CptN(CurrLayer) = cNone

    DEFINT xLen = EdgeBMP.Width \ 30                                    'sets resolution of control points to search
    DEFINT xLmit = EdgeBMP.Width -2
    DEFINT MetXlimit = False
    x = 0
    DO                                                                  'march from left to right
        IF x = xLmit THEN MetXlimit = True                              'at end, do one more last point
        DO
            IF (&HFF AND EdgeBMP.pixel(x, y)) THEN                      'find ILM edge from smaller EdgeBMP
                xd = x * Scale_Factor
                yd = (y - SearchRes) * Scale_Factor                     'back up some
                yLmit = (y + SearchRes) * Scale_Factor                  'finish setting search range
                j = FindMaxVertChange(SegmentBMP, xd, yd, yLmit)        'do fine search on pre-filtered OCT image, faster than FindMaxEdgeChange
                yd = IIF( (j = NullResult), y * Scale_Factor, j)        'get something?
                xd = xd + OCT_Left
                AddControlPoint(xd, yd)                                 'updates array CptX, CptY, CptN
                y = y - MaxILMChange                                    'go back up ~ XX * Scale_Factor * 9 (eg, 40 = 360) microns for non-flattened OCT
                IF y < StrtY THEN y = StrtY
                EXIT DO
            ELSE
                y ++
            END IF
        LOOP UNTIL (y = EndY)
        IF (y = EndY) THEN y = StrtY                                    'nothing found
              
        IF MetXlimit = False THEN                                       'still marching
            INC(x, xLen)
            IF x > EdgeBMP.Width THEN x = xLmit                         'finished marching across for ILM
        END IF                                                          'next go around will break loop
    LOOP UNTIL MetXlimit

    RemoveControlPointOutliers(CurrLayer, MaxILMChange * 0.5)

    RESULT = CptN(CurrLayer)                                            'return # control points found, x,y locations are temp stored in CptX, CptY
END FUNCTION





'========================================================================
'RNFL segmentation uses ILM and ONL  segmentations for range of search 
'========================================================================

FUNCTION SegmentRNFL() AS INTEGER
    CurrLayer = cNFL                                                    'setting RNFL layer
    CptN(CurrLayer) = cNone
    DEFINT ndx = 0
    DEFINT j, k, xd, yd, yLmit                                          'bottom x,y search range
    DEFINT MaxRNFLRange = 85 \ Scale_Factor                             'maximum thickness of RNFL 85 * xx microns axial res ~ 4 * 85 = XXX microns
    DEFINT RNFLOffset = 4                                               'offset from ILM to start search

    DO                                                                  'running across horizontal points

        xd = (CptX(cILM, ndx) - OCT_Left) \ Scale_Factor                'X position search for end of search
        yd = (CptY(cILM, ndx) \ Scale_Factor) + RNFLOffset              'Y position search for end of search

        IF (VesselMask.Pixel(xd, yd) = 0) THEN                         'no vessel shadow detected in this vertical column
            yLmit = yd + MaxRNFLRange
            k = FindMinOnOCT(kMeanCopy, xd, yLmit, yd)                  'should be OPL/ONL border
            j = FindChangeOnOCT(kMeanCopy, xd, yd, k-2)                 'search from ILM to ONL
            IF (j = NullResult) OR (j >= k) THEN j = yLmit
            IF j = yd THEN j = yLmit
            IF j <= yd THEN j = yd + 1                                  'RNFL too thin? Get single pixel diff

            yd = (j - RNFLOffset) * Scale_Factor                        'back up some for fine tuning
            yLmit = (j + RNFLOffset) * Scale_Factor                     'forward some for fine tuning
            xd = xd * Scale_Factor                                      'convert back from small to large
            j = FindDecreaseOnOCT(SegmentBMP, xd, yd, yLmit, 0.9)       'do fine search on full pre-filtered OCT image, somewhat better than FindMaxVertChange(
            yd = IIF( (j <> NullResult), j, (yd + yLmit)\2)             'get something?
            xd = xd + OCT_Left                                          'convert back to form coordinates
            AddControlPoint(xd, yd)                                     'updates array CptX, CptY, CptN
        END IF

        ndx ++
    LOOP UNTIL (ndx > CptN(cILM))                                       'finished using ILM points as reference

    RemoveControlPointOutliers(CurrLayer, MaxRNFLRange * Scale_Factor * 0.5)

    RESULT = CptN(CurrLayer)                                            'return # control points found, actual x,y locations are temp stored in SegX, SegY
END FUNCTION





'========================================================================
' Outer Nuclear layer segmentation assumes Edge BMP has dark strip
'  between ILM and Bruch's membrane. Should be even for low reflectance
'========================================================================

FUNCTION SegmentONL() AS INTEGER

    CurrLayer = cONL                                                    'setting top of ONL 
    CptN(CurrLayer) = cNone
    DEFINT ndx = 0
    DEFINT i , j                                                        'loopers
    DEFINT xd , yd , yLmit                                              'temp vars for looping and rescaling
    DEFINT ONLOffset = 2                                                'starting offset from NFL to search for ONL
    DEFINT MaxONLLRange = 40 \ Scale_Factor                             'maximum thickness of RNFL 85 * xx microns axial res ~ 4 * 85 = 340 microns

    DO                                                                  'for running horizontally across ILM segmentation points
        xd = (CptX(cNFL, ndx) - OCT_Left) \ Scale_Factor                'X position search for end of search
        yd = (CptY(cNFL, ndx)\Scale_Factor) + ONLOffset                 'start Y position search
        yLmit = yd + MaxONLLRange                                       'limit ONL search

        IF (VesselMask.Pixel(xd, yd) = 0) THEN                          'no vessel shadow detected in this vertical column
            j = FindMinOnOCT(kMeanCopy, xd, yd, yLmit)                  'should be OPL/ONL border, use RetinaMask?; better than FindMaxDecreaseOnOCT(SegmentBMP, xd, yd, yLmit)
            xd = xd * Scale_Factor 
            IF (j <> NullResult) THEN                                   'fine tune
                'IF j = yd THEN j = yd + 1                               'too thin? Get single pixel diff
                yd = (j * Scale_Factor) - Scale_Factor
                yLmit = yd + Scale_Factor
                j = FindMaxVertChange(SegmentBMP, xd, yd, yLmit)
            ELSE
                yd = (yd + ONLOffset) * Scale_Factor
            END IF

            IF (j <> NullResult) THEN
                yd = j + 2                                              'empirical offset
            END IF

            xd = xd + OCT_Left                                          'convert back from small to large, then format for form coordinates
            AddControlPoint(xd, yd)                                     'updates array CptX, CptY, CptN
        END IF

        ndx ++
    LOOP UNTIL (ndx > CptN(cNFL))                                        'only using ILM points as reference

    '--------  Perform polynomial fit and correct outliers  -------------
    RemoveControlPointOutliers(CurrLayer, MaxONLLRange * Scale_Factor * 0.5)
    IF PolyFitLayersCheckBox.Checked THEN CorrectLayerOutliers(CurrLayer, VAL(PolyFitLayersOrderBox.Item(PolyFitLayersOrderBox.ItemIndex)), 5)

    RESULT = CptN(CurrLayer)                            'return # control points found, actual x,y locations are temp stored in SegX, SegY
END FUNCTION




'========================================================================
'now do ellipsoid zone segmentation
'========================================================================
'FUNCTION SegmentEZL() AS INTEGER
'    DEFINT ndx = 0
'    DEFINT i , j                                                        'loopers
'    DEFINT xd , yd, yLmit                                               'temp vars for searching
'    DEFINT x , y                                                        'coordinates on edge filtered image
'    DEFINT MaxEZLOffset= 40 \ Scale_Factor
'
'    CurrLayer = cEZL                                                    'setting EZ layer edge
'    CptN(CurrLayer) = cNone
'
'    DO                                                                  'for running horizontally across segmentation points
'        xd = (CptX(cONL, ndx) - OCT_Left) \ Scale_Factor                'X position search for end of search
'        yd = (CptY(cONL, ndx) \ Scale_Factor) + 4                       'start Y position with offset, 'ONL already checked for vessel shadow
'        yLmit = yd + MaxEZLOffset                                       'limit  search based on retinal thickness
'        'IF UseKmeansCheckBox.Checked THEN
'            j = FindChangeOnOCT(kMeanCopy, xd, yd, yLmit)
'        'ELSE
'        '    j = FindMaxVertChange(RetinaMask, xd, yd, yLmit)
'        'END IF
'        yd = IIF( (j <> NullResult), j * Scale_Factor, yLmit * Scale_Factor)
'        xd = (xd * Scale_Factor) + OCT_Left                                      'convert back to full size and form coordinates
'        AddControlPoint(xd, yd)                                 'updates array CptX, CptY, CptN
'        SegX(CurrLayer , SegN(CurrLayer)) = xd                  'x coordinate of control points of line making up a segmentation series
'        SegY(CurrLayer , SegN(CurrLayer)) = yd                  'corrected y coordinate of control points of line making up a segmentation series
'        INC(SegN(CurrLayer), 1)                                 'total number of points for a segmentation line for each series
'        IF SegN(CurrLayer) > MaxPts THEN SegN(CurrLayer) = MaxPts
'
'        ndx ++
'    LOOP UNTIL (ndx > CptN(cONL))                               'finished marching across for BM
'
'    '--------  Perform polynomial fit and correct outliers  -------------
'    RemoveControlPointOutliers(CurrLayer, MaxEZLOffset * Scale_Factor)
'    IF PolyFitLayersCheckBox.Checked THEN CorrectLayerOutliers(CurrLayer, VAL(PolyFitLayersOrderBox.Item(PolyFitLayersOrderBox.ItemIndex)), 7)
'
'    RESULT = CptN(CurrLayer)                                'return # control points found, actual x,y locations are temp stored in SegX, SegY
'END FUNCTION





'========================================================================
'Bruch's membrane segmentation using EZ layer segmentation
'searchs SegmentBMP based on EZ layer
'========================================================================

FUNCTION SegmentBML() AS INTEGER
    DEFINT ndx = 0
    DEFINT j, xd , yd, yLmit                                            'temp vars for searching
    DEFINT BMStartOffset = 4
    DEFINT MaxBMOffset= 40 \ Scale_Factor                               'working offset from ONL to BM layer (XX * 9 microns)
    DEFINT MaxEZLThickness = 16                                         'working offset from EZL to bottom of BM layer (XX * 9 microns) full scale

    CurrLayer = cBM                                                     'setting Bruch's membrane layer edge
    CptN(CurrLayer) = cNone


    DO                                                                  'for running horizontally across segmentation points
        xd = (CptX(cONL, ndx) - OCT_Left) \ Scale_Factor                'X position search for end of search
        yd = (CptY(cONL, ndx) \ Scale_Factor) + BMStartOffset           'start Y position with offset, 'ONL already checked for vessel shadow
        yLmit = yd + MaxBMOffset                                        'limit  search based on retinal thickness

        j = FindMaxOnOCT(RetinaMask, xd, yd, yLmit)                     'tends to get interdigitation zone

        IF (j <> NullResult) THEN
            xd = xd * Scale_Factor
            yd = (j + 2) * Scale_Factor
            j = FindMaxOnOCT(SegmentBMP, xd, yd + MaxEZLThickness, yd)
            yd = IIF ((j <> NullResult), j + 2, yLmit * Scale_Factor)
        ELSE
            xd = xd * Scale_Factor
            yd = yd * Scale_Factor + 2
        END IF

        xd = xd + OCT_Left
        AddControlPoint(xd, yd)                                 'updates array CptX, CptY, CptN
        ndx ++
    LOOP UNTIL (ndx > CptN(cONL))

    '--------  Perform polynomial fit and correct outliers  -------------
    RemoveControlPointOutliers(CurrLayer, MaxEZLThickness * Scale_Factor * 0.5)
    IF PolyFitLayersCheckBox.Checked THEN CorrectLayerOutliers(CurrLayer, VAL(PolyFitLayersOrderBox.Item(PolyFitLayersOrderBox.ItemIndex)), 5)

    RESULT = CptN(CurrLayer)                        'return # control points found, actual x,y locations are temp stored in SegX, SegY
END FUNCTION






'========================================================================
'Choroid segmentation using EZ layer segmentation
'========================================================================

FUNCTION SegmentChoroid() AS INTEGER
    DEFINT TheEdgeThick = 12                        'set thickness of edges after sobel
    DEFINT xLen = 10                                'sets resolution of control points to search
    DEFINT EndY = TheEdgeThick                      'start top vertical offset
    DEFINT StrtY = EdgeBMP.Height - TheEdgeThick    'end of vertical search
    DEFINT i , j                                    'loopers
    DEFINT xd , yd, yLmit                           'temp vars for searching
    DEFINT x , y = TheEdgeThick * 2                 'coordinates on edge filtered image
    DEFINT MaxBMThickness = 9

    CurrLayer = cBM                                 'setting Bruch's membrane layer edge
    CptN(CurrLayer) = cNone

    FOR x = xLen TO(RetinaMask.Width - xLen) STEP xLen                     'march left to right
        y = StrtY
        DO
            IF RetinaMask.pixel(x, y) THEN                                 'find BM edge from scaled original  ...
                j = FindMaxOnOCT(SmallCopy, x, y, y-MaxBMThickness)
                xd = (x * Scale_Factor) + OCT_Left
                yd = IIF( (j <> NullResult), (j+Scale_Factor) * Scale_Factor, y * Scale_Factor)
                AddControlPoint(xd, yd)                                 'will update array CptX, CptY, CptN
                y = y + 20                                              'go back up ~ 30 * Scale_Factor * 9 (180) microns for non-flattened OCT
                IF y > StrtY THEN y = StrtY
                EXIT DO
            ELSE
                y --
            END IF
        LOOP UNTIL (y = EndY)
        IF (y = EndY) THEN y = StrtY            'nothing found
    NEXT x                                      'finished marching across for ILM


    RESULT = CptN(CurrLayer)        'return # control points found, actual x,y locations are temp stored in SegX, SegY
END FUNCTION







FUNCTION CorrectLayerOutliers(TheLayer AS INTEGER, PolyOrderNum AS INTEGER, TheCutOff AS INTEGER) AS INTEGER

    DEFINT i, n = 0
    DEFINT ArryElementNum = CptN(TheLayer)

    FOR i = 0 TO ArryElementNum
        InFitDataX(i) = CptX(TheLayer, i)
        InFitDataY(i) = CptY(TheLayer, i)
    NEXT i


    IF FitCurve (InFitDataX(), InFitDataY(), ArryElementNum, PolyOrderNum, OutputCoeff()) <> False THEN
        DEFINT xpoint, ypoint

        FOR i = 0 TO ArryElementNum
            ypoint = CEIL(GetYprediction(CptX(TheLayer, i), PolyOrderNum, OutputCoeff()))
            IF ABS(CptY(TheLayer, i) - ypoint) > TheCutOff THEN
                CptY(TheLayer, i) = ypoint
                n ++
            END IF
        NEXT i
    END IF
    RESULT = n
END FUNCTION



'
'SUB FlattenOCT()
'
'    DEFINT ArryElementNum
'    DEFINT MaxY, i
'    DEFINT xpoint, ypoint
'    DEFINT PolyOrderNum = 3'VAL(PolyFitOrderEdit.Text)
'        IF PolyOrderNum > MaxPolyOrder THEN PolyOrderNum = MaxPolyOrder   ': PolyFitOrderEdit.Text = STR$(PolyOrderNum)
'        IF PolyOrderNum < 2            THEN PolyOrderNum = 2              ': PolyFitOrderEdit.Text = STR$(PolyOrderNum)
'
'    FlattenStripBMP.Height = OCTimg.Height
'    FlattenStripBMP.Width = 1
'    DIM Dest AS QRECT
'        Dest.Top = 0
'        Dest.Left = 0
'        Dest.Bottom = FlattenStripBMP.Height
'        Dest.Right = FlattenStripBMP.Width
'
'
'        FilterOCTForSegmentation
'        SegmentBM
'
'        IF CptN(cBM) > (PolyOrderNum + 1) THEN          'need a minimum # of data points from Bruch's membrane layer for fit
'
'            '--------  Perform polynomial fit  Bruch's membrane layer edge   -------------
'            ArryElementNum = CptN(cBM)
'            MaxY = NullResult
'            FOR i = 0 TO ArryElementNum
'                InFitDataX(i) = CptX(cBM, i)
'                InFitDataY(i) = CptY(cBM, i)
'                IF (InFitDataY(i) > MaxY) THEN MaxY = InFitDataY(i)
'            NEXT i
'
'            IF FitCurve (InFitDataX(), InFitDataY(), ArryElementNum, PolyOrderNum, OutputCoeff()) <> False THEN
'                '--------------------------------------------
'                'get results here
'                FOR xpoint = 0 to OCTimg.Width
'                    ypoint = CEIL(GetYprediction((xpoint * 1.000), PolyOrderNum, OutputCoeff()))
'                    ypoint = MaxY - ypoint
'                    myRect.Left = xpoint
'                    myRect.Right = xpoint + 1
'                    myRect.Top = Dest.Top      
'                    myRect.Bottom = Dest.Bottom             'all points need refreshing!!!
'                    FlattenStripBMP.CopyRect(Dest, OCTimg, myRect)
'                    OCTimg.Line(xpoint, 0, xpoint,  OCTimg.Height, clBlack)
'                    OCTimg.Draw(xpoint, ypoint, FlattenStripBMP.BMP)
'                NEXT xpoint
'            END IF
'        END IF
'
'
'
'    FileLoadGauge.Visible = False
'    CancelLoadButton.Visible = False
'    Trackbar_Change
'
'END SUB





'----------------------------------------------------------------
'uses global EdgeFilterX and EdgeFilterY
'for kernel coordinates to sum across edge
'----------------------------------------------------------------



FUNCTION FindDecreaseOnOCT(TheBMP AS QBITMAP , inX AS INTEGER , inY AS INTEGER , EndY AS INTEGER , Thresh AS SINGLE) AS INTEGER
    DEFINT i , y
    DEFINT TheSgn = IIF( (inY < EndY), 1, -1)
    DEFINT Rtn = NullResult
    RESULT = Rtn  'set for now

    DEFDBL SumLocal = 0
    DEFDBL BaseIntensity
    i = 0
    DO  'march vertically summing edge filter detectors
        INC(SumLocal, &HFF AND TheBMP.pixel(inX + EdgeFilterX(i) , inY - (TheSgn * EdgeFilterY(i))))
        i ++
    LOOP UNTIL (i > NumEdgeFilterElements)

    BaseIntensity = SumLocal
    IF BaseIntensity < 1 THEN BaseIntensity = 1
    y = inY

    DO
        SumLocal = 0.0
        i = 0
        DO  'march vertically summing edge filter detectors
            INC(SumLocal, &HFF AND TheBMP.pixel(inX + EdgeFilterX(i) , y + (TheSgn * EdgeFilterY(i))))
            i ++
        LOOP UNTIL (i > NumEdgeFilterElements)

        IF (SumLocal / BaseIntensity) < Thresh THEN
            Rtn = y 
            EXIT DO
        END IF
        y = y + TheSgn
    LOOP UNTIL (y = EndY)

    RESULT = Rtn
END FUNCTION



FUNCTION FindIncreaseOnOCT(TheBMP AS QBITMAP , inX AS INTEGER , inY AS INTEGER , EndY AS INTEGER , Thresh AS SINGLE) AS INTEGER
    DEFINT i , y
    DEFDBL SumLocal = 0
    DEFDBL BaseIntensity
    DEFINT TheSgn = IIF( (inY < EndY), 1, -1)
    DEFSNG Rtn = NullResult
    RESULT = Rtn  'set for now

    i = 0
    DO  'march vertically summing edge filter detectors
        INC(SumLocal, &HFF AND TheBMP.pixel(inX + EdgeFilterX(i) , inY - (TheSgn * EdgeFilterY(i))))
        i ++
    LOOP UNTIL (i > NumEdgeFilterElements)

    BaseIntensity = SumLocal
    IF BaseIntensity < 1 THEN BaseIntensity = 1
    y = inY

    DO
        SumLocal = 0.0
        i = 0
        DO  'march vertically summing edge filter detectors
            INC(SumLocal, &HFF AND TheBMP.pixel(inX + EdgeFilterX(i) , y + (TheSgn * EdgeFilterY(i))))
            i ++
        LOOP UNTIL (i > NumEdgeFilterElements)

        IF (SumLocal / BaseIntensity) > Thresh THEN
            Rtn = y 
            EXIT DO
        END IF
        y = y + TheSgn
    LOOP UNTIL (y = EndY)

    RESULT = Rtn
END FUNCTION





FUNCTION FindMaxIncreaseOnOCT(TheBMP AS QBITMAP , inX AS INTEGER , inY AS INTEGER , EndY AS INTEGER) AS INTEGER
    DEFINT i , y
    DEFDBL SumLocal = 0
    DEFDBL BaseIntensity
    DEFDBL MaxChange = 99999
    DEFSNG Rtn = NullResult
    RESULT = Rtn  'set for now

    i = 0
    DO  'march vertically summing edge filter detectors
        INC(SumLocal , &HFF AND TheBMP.pixel(inX + EdgeFilterX(i) , inY - EdgeFilterY(i)))
        i ++
    LOOP UNTIL (i > NumEdgeFilterElements)

    BaseIntensity = SumLocal
    IF BaseIntensity < 1 THEN BaseIntensity = 1
    y = inY

    DO
        SumLocal = 0.0
        i = 0
        DO  'march vertically summing edge filter detectors
            INC(SumLocal , &HFF AND TheBMP.pixel(inX + EdgeFilterX(i) , y + EdgeFilterY(i)))
            i ++
        LOOP UNTIL (i > NumEdgeFilterElements)

        IF (SumLocal / BaseIntensity) > MaxChange THEN
            MaxChange = (SumLocal / BaseIntensity)
            Rtn = y
        END IF
        y ++
    LOOP UNTIL (y > EndY)

    RESULT = Rtn
END FUNCTION





FUNCTION FindMaxDecreaseOnOCT(TheBMP AS QBITMAP , inX AS INTEGER , inY AS INTEGER , EndY AS INTEGER) AS INTEGER
    DEFINT i , y
    DEFDBL SumLocal = 0
    DEFDBL BaseIntensity
    DEFDBL MaxChange = 99999
    DEFSNG Rtn = NullResult
    RESULT = Rtn  'set for now

    i = 0
    DO  'march vertically summing edge filter detectors
        INC(SumLocal , &HFF AND TheBMP.pixel(inX + EdgeFilterX(i) , inY - EdgeFilterY(i)))
        i ++
    LOOP UNTIL (i > NumEdgeFilterElements)

    BaseIntensity = SumLocal
    IF BaseIntensity < 1 THEN BaseIntensity = 1
    y = inY

    DO
        SumLocal = 0.0
        i = 0
        DO  'march vertically summing edge filter detectors
            INC(SumLocal , &HFF AND TheBMP.pixel(inX + EdgeFilterX(i) , y + EdgeFilterY(i)))
            i ++
        LOOP UNTIL (i > NumEdgeFilterElements)

        IF (SumLocal / BaseIntensity) < MaxChange THEN
            MaxChange = (SumLocal / BaseIntensity)
            Rtn = y
        END IF
        y ++
    LOOP UNTIL (y > EndY)

    RESULT = Rtn
END FUNCTION





FUNCTION FindLocalMin(TheBMP AS QBITMAP , inX AS INTEGER , inY AS INTEGER , EndY AS INTEGER, RangeY AS INTEGER) AS INTEGER
    DEFINT Upp3, Upp2, Upp1, Midl, Lowr1, Lowr2
    DEFINT Rtn = NullResult
    DEFINT i, y

    y = inY

    DO
        i = 0
        Upp3 = 0
        Upp2 = 0
        Upp1 = 0
        Midl = 0
        Lowr1 = 0
        Lowr2 = 0

        DO  'march vertically summing edge filter detectors
            IF RangeY > 0 THEN
                INC(Upp3 ,  &HFF AND TheBMP.pixel(inX + EdgeFilterX(i), y + EdgeFilterY(i)) -3)
                INC(Upp2 ,  &HFF AND TheBMP.pixel(inX + EdgeFilterX(i), y + EdgeFilterY(i)) -2)
                INC(Lowr2 , &HFF AND TheBMP.pixel(inX + EdgeFilterX(i), y + EdgeFilterY(i)) +2)
            END IF

            INC(Upp1 ,  &HFF AND TheBMP.pixel(inX + EdgeFilterX(i), y + EdgeFilterY(i)) -1)
            INC(Midl ,  &HFF AND TheBMP.pixel(inX + EdgeFilterX(i), y + EdgeFilterY(i))   )
            INC(Lowr1 , &HFF AND TheBMP.pixel(inX + EdgeFilterX(i), y + EdgeFilterY(i)) +1)
            i ++
        LOOP UNTIL (i > NumEdgeFilterElements)

        if Upp2 < 1 THEN Upp2 = 1
        if Upp1 < 1 THEN Upp1 = 1
        if Midl < 1 THEN Midl = 1
        IF RangeY > 0 THEN
                IF (Upp1 > Midl) AND (Lowr1 >= Midl) THEN
                    Rtn = i
                    EXIT DO
                END IF
        ELSE
                IF ((Upp3/Upp2) > 1.25) AND ((Upp2/Upp1) > 1.25) AND ((Upp1 / Midl) > 1.0)  AND (Lowr1 >= Midl) THEN 'AND (Lowr2 >= Midl) THEN
                    Rtn = i
                    EXIT DO
                END IF
        END IF
        y ++
    LOOP UNTIL (y > EndY)
    RESULT = Rtn
END FUNCTION




FUNCTION FindMaxVertChange(TheBMP AS QBITMAP , inX AS INTEGER , inY AS INTEGER , EndY AS INTEGER) AS INTEGER
    DEFINT SumFirst, SumNext
    DEFINT TheSgn = IIF( (inY < EndY), 1, -1)
    DEFINT Rtn = NullResult
    DEFINT i, y = inY 
    DEFINT maxDelta = 0

    DO
        i = 0
        SumFirst = 0
        SumNext = 0
        DO 
            INC(SumFirst, &HFF AND TheBMP.Pixel(inX + VertFilterX(i), y + VertFilterY(i)         ))
            INC(SumNext,  &HFF AND TheBMP.Pixel(inX + VertFilterX(i), y + VertFilterY(i) + TheSgn))
            i ++
        LOOP UNTIL (i > NumVertFilterElements)
        IF ABS(SumFirst - SumNext) > maxDelta THEN
            maxDelta = ABS(SumFirst - SumNext)
            Rtn = y
        END IF
        y = y + TheSgn
    LOOP UNTIL ( y = EndY)

    RESULT = Rtn
END FUNCTION



FUNCTION FindMaxEdgeChange(TheBMP AS QBITMAP , inX AS INTEGER , inY AS INTEGER , EndY AS INTEGER) AS INTEGER
    DEFINT SumFirst, SumNext 
    DEFINT TheSgn = IIF( (inY < EndY), 1, -1)
    DEFINT Rtn = NullResult
    DEFINT i, y = inY 
    DEFINT maxDelta = 0

    DO
        i = 0
        SumFirst = 0
        SumNext = 0
        DO 
            INC(SumFirst, &HFF AND TheBMP.Pixel(inX + EdgeFilterX(i), y + EdgeFilterY(i)         ))
            INC(SumNext,  &HFF AND TheBMP.Pixel(inX + EdgeFilterX(i), y + EdgeFilterY(i) + TheSgn))
            i ++
        LOOP UNTIL (i > NumEdgeFilterElements)
        IF (ABS(SumFirst - SumNext ) > maxDelta) THEN
            maxDelta = ABS(SumFirst - SumNext)
            Rtn = y
        END IF
        y = y + TheSgn
    LOOP UNTIL ( y = EndY)

    RESULT = Rtn
END FUNCTION



FUNCTION FindZeroOnOCT(TheBMP AS QBITMAP , inX AS INTEGER , inY AS INTEGER, EndY AS INTEGER) AS INTEGER
    DEFINT Rtn = NullResult
    DEFINT i = inY
    DEFINT TheSgn = IIF( (inY < EndY), 1, -1)

    DO
        IF TheBMP.pixel(inX , i) = 0 THEN
            Rtn = i
            EXIT DO
        END IF
        i = i + TheSgn
    LOOP UNTIL (i = endY)

    RESULT = Rtn
END FUNCTION





FUNCTION FindMinOnOCT(TheBMP AS QBITMAP , inX AS INTEGER , inY AS INTEGER , EndY AS INTEGER) AS INTEGER
    DEFINT ThePeak = 99999
    DEFINT TheSgn = IIF( (inY < EndY), 1, -1)
    DEFINT Rtn = NullResult
    DEFINT i, SumLocal
    DEFINT y = inY

    DO
        i = 0
        SumLocal = 0
        DO 
            INC(SumLocal, &HFF AND TheBMP.pixel(inX + VertFilterX(i) , y + (TheSgn * VertFilterY(i))))
            i ++
        LOOP UNTIL (i > NumVertFilterElements)


        IF SumLocal < ThePeak THEN
            ThePeak = SumLocal
            Rtn = y
        END IF
        y = y + TheSgn
    LOOP UNTIL (y = EndY)

    RESULT = Rtn
END FUNCTION



FUNCTION FindMaxOnOCT(TheBMP AS QBITMAP , inX AS INTEGER , inY AS INTEGER , EndY AS INTEGER) AS INTEGER
    DEFINT ThePeak = NullResult
    DEFINT TheSgn = IIF( (inY < EndY), 1, -1)
    DEFINT Rtn = NullResult
    DEFINT i, SumLocal
    DEFINT y = inY

    DO
        i = 0
        SumLocal = 0
        DO 
            INC(SumLocal, &HFF AND TheBMP.pixel(inX + VertFilterX(i) , y + (TheSgn * VertFilterY(i))))
            i ++
        LOOP UNTIL (i > NumVertFilterElements)

        IF SumLocal > ThePeak THEN
            ThePeak = SumLocal
            Rtn = y
        END IF
        y = y + TheSgn
    LOOP UNTIL (y = EndY)

    RESULT = Rtn
END FUNCTION





FUNCTION FindChangeOnOCT(TheBMP AS QBITMAP , inX AS INTEGER , inY AS INTEGER , EndY AS INTEGER) AS INTEGER
    DEFINT TheSgn = IIF( (inY < EndY), 1, -1)
    DEFINT Rtn = NullResult
    DEFINT y = inY
    DEFINT Baseline = TheBMP.pixel(inX, y)
    DO
        y = y + TheSgn
        IF (TheBMP.pixel(inX, y) <>  Baseline) AND (TheBMP.pixel(inX, y+1) <>  Baseline) AND (TheBMP.pixel(inX, y+2) <>  Baseline) THEN
            Rtn = y
            EXIT DO
        END IF
    LOOP UNTIL (y = EndY)

    RESULT = Rtn
END FUNCTION




