
'------------------------------------------------------------
'       Catmull-Rom spline fitting routines
'       This module should stand on its own only requiring globals
'       MaxSlices, MaxLayers, MaxPts, MaxCtrlPts
'------------------------------------------------------------


DIM SegX(MaxLayers, MaxPts) AS INTEGER      'x points of line making up a segmentation series
DIM SegY(MaxLayers, MaxPts) AS INTEGER      'y points of line making up a segmentation series
DIM SegN(MaxLayers)         AS INTEGER      'total number of points making up a segmentation line for each series

DIM SegColor(MaxLayers)     AS INTEGER      'line color for each segment
    SegColor(cILM) = RGB(0,   64, 255)
    SegColor(cNFL) = RGB(0,   255, 0)
    SegColor(cONL) = RGB(0,   255, 255)
    'SegColor(cEZL) = RGB(255, 255, 0)
    SegColor(cBM) =  RGB(255, 0  , 255)


DEFINT OrigX, OrigY                         'globals for dragging locations
DEFINT CtrlPtDrag                           'flag dragging a control point
DEFINT CtrlPtHalfHeight = 4                 'half height for centering
DEFINT CtrlPtHalfWidth = 4                  'half width for centering
DEFINT SplineLocked(MaxLayers)              'finished storing a drawn spline
       LockAllSplines(False)                'init

DEFINT CtrlKeyIsDown                        'Ctrl Key active, allows circle outline shift of control points
DEFINT ControlPointCursorSGN = NullResult   'Vertical Orientation sign onKeyDown when users allows  allows circle outline shift of control points
DEFINT DrawFlag = False                     'starting with mouse down, user is currently drawing a spline
DEFINT CurrLayer                            'current selected layer segmentation
DEFSNG SplineGap = 4.00                     'ratio of OCT pixels to spline points, use large values for fast editing, use 1 for final spline fit



'------------------------------------------------------------
' Control points for splines
'------------------------------------------------------------
DIM CtrlPt(MaxCtrlPts)          AS QPANEL   '(array of QPanel) for active editing, can be use for either internal limiting membrane or Nerve fiber layer
DIM CptX(MaxLayers, MaxCtrlPts) AS INTEGER  'x coordinate of control points of line making up a segmentation series
DIM CptY(MaxLayers, MaxCtrlPts) AS INTEGER  'y coordinate of control points of line making up a segmentation series
DIM CptN(MaxLayers)             AS INTEGER  'total number of control points for controllnig a segmentation line for each series
    CptN(0) = -1                            'initialize to none (-1)
    CptN(1) = -1




'************************************************************************
' Initialize control point events
' since all are the same
'************************************************************************
FOR CtrlPtDrag = 0 TO MaxCtrlPts
    CtrlPt(CtrlPtDrag).Visible = False
    CtrlPt(CtrlPtDrag).OnMouseDown = CtrlPtMouseDown
    CtrlPt(CtrlPtDrag).OnMouseUp = CtrlPtMouseUp
    CtrlPt(CtrlPtDrag).OnMouseMove = CtrlPtMouseMove
    CtrlPt(CtrlPtDrag).OnDblClick = InsertControlPoint
    CtrlPt(CtrlPtDrag).Cursor = crhandpoint
NEXT CtrlPtDrag




SUB AddControlPoint(x AS INTEGER, y AS INTEGER)
    CtrlPtDrag = False
    IF CptN(CurrLayer) = MaxCtrlPts THEN  EXIT SUB
    INC(CptN(CurrLayer), 1)
    DEFINT n = CptN(CurrLayer)                  'easier to read

    CtrlPt(n).Parent = Form
    CtrlPt(n).Color = clRed
    CtrlPt(n).BevelWidth = 0
    CtrlPt(n).Width = CtrlPtHalfWidth * 2
    CtrlPt(n).Height = CtrlPtHalfHeight * 2
    CtrlPt(n).Left = x - CtrlPtHalfWidth
    CtrlPt(n).Top = y - CtrlPtHalfHeight
    CtrlPt(n).Tag = n                           'keep index for events
    CtrlPt(n).Enabled = True

    'copy coordinates for switching views
    CptX(CurrLayer, n) = x 
    CptY(CurrLayer, n) = y 
END SUB





SUB InsertControlPoint(Sender AS QPANEL)
    CtrlPtDrag = False

    IF CptN(CurrLayer) = MaxCtrlPts THEN EXIT SUB                               'all used up
    IF Sender.Tag = CptN(CurrLayer) THEN EXIT SUB                               'can't insert on last

    DEFINT NewX = (CtrlPt(Sender.Tag).Left + CtrlPt(Sender.Tag + 1).Left) / 2 'CtrlPt(Sender.Tag).Width + CtrlPtHalfWidth                 'location of new one
    DEFINT NewY = (CtrlPt(Sender.Tag).Top  + CtrlPt(Sender.Tag + 1).Top) / 2 'CtrlPt(Sender.Tag).Height + CtrlPtHalfHeight

    'add new control point at last one, then fill in the middle later
    AddControlPoint(CtrlPt(CptN(CurrLayer)).Left + CtrlPtHalfWidth, _
                    CtrlPt(CptN(CurrLayer)).Top  + CtrlPtHalfHeight)

    DEFINT n = CptN(CurrLayer)                                                  'modified by adding control point, easier to read
    DEFINT k

    FOR k = (n-1) TO (Sender.Tag+1) STEP -1                                     'start from last to first!!
        CtrlPt(k).Left = CtrlPt(k-1).Left                                       'shift up in index
        CtrlPt(k).Top = CtrlPt(k-1).Top 
        CtrlPt(k).Tag = k
        CptX(CurrLayer, k) = CtrlPt(k).Left + CtrlPtHalfWidth                   'copy for each series using qpanels
        CptY(CurrLayer, k) = CtrlPt(k).Top + CtrlPtHalfHeight
    NEXT k

    'finish up
    FOR k = 0 TO n
        CtrlPt(k).Visible = True                        'do this just in case dblClick BEFORE activating a layer
    NEXT k
    CtrlPt(Sender.Tag+1).Left = NewX
    CtrlPt(Sender.Tag+1).Top  = NewY
    CtrlPt(Sender.Tag+1).Tag  = Sender.Tag+1
    CptX(CurrLayer, Sender.Tag+1) = NewX                'copy for each series using qpanels
    CptY(CurrLayer, Sender.Tag+1) = NewY
    CtrlPtDrag = False
END SUB





SUB DeleteControlPoint(NumCtrlPt AS INTEGER) 
    CtrlPtDrag = False
    IF CptN(CurrLayer) < 4 THEN EXIT SUB                'need at least 3 control points
    IF NumCtrlPt = CptN(CurrLayer) THEN                 'delete the last one
        CtrlPt(CptN(CurrLayer)).Visible = False
        'CtrlPt(CptN(CurrLayer)).Enabled = False
        CptN(CurrLayer) = CptN(CurrLayer) -1
        EXIT SUB
    END IF

    DEFINT k
    FOR k = NumCtrlPt TO CptN(CurrLayer)-1
        CtrlPt(k).Left = CtrlPt(k+1).Left               'shift down in index
        CtrlPt(k).Top = CtrlPt(k+1).Top 
        CtrlPt(k).Tag = k
        CptX(CurrLayer, k) = CtrlPt(k+1).Left + CtrlPtHalfWidth     'copy for each series using qpanels
        CptY(CurrLayer, k) = CtrlPt(k+1).Top  + CtrlPtHalfHeight
    NEXT k
    CtrlPt(CptN(CurrLayer)).Visible = False
    'CtrlPt(CptN(CurrLayer)).Enabled = False
    CptN(CurrLayer) = CptN(CurrLayer) -1
END SUB




'--------------------------------------------------------------------
'   double-click inserts a control point nearest the cursor
'--------------------------------------------------------------------

SUB AddControlPointbyDblClick(x AS INTEGER , y AS INTEGER)
    DEFINT k

    FOR k = 0 TO CptN(CurrLayer)-1
        IF (x > CtrlPt(k).Left) AND (x <= CtrlPt(k+1).Left) THEN                    'find corresponding control point to the LEFT..
            InsertControlPoint(CtrlPt(k))                                           'need Insert for proper sequence, NOT AddControlPoint(x, y) this adds at end...
            CtrlPt(k+1).Top = y - CtrlPtHalfHeight                                  'now reposition control point near the click location
            CtrlPt(k+1).Left = x - CtrlPtHalfWidth
            CptX(CurrLayer, k+1) = CtrlPt(k+1).Left + CtrlPtHalfWidth               'reposition coordinates for spline fitting
            CptY(CurrLayer, k+1) = CtrlPt(k+1).Top  + CtrlPtHalfHeight
            EXIT FOR
        END IF
    NEXT k

END SUB





SUB RemoveControlPointOutliers(TheLayer AS INTEGER, TheMaxChange AS INTEGER)

    IF CptN(TheLayer) < 3 THEN EXIT SUB                                             'not enough data

    DEFINT k = 0

    'Check first point to adjacent two
    IF CptX(TheLayer, k) > CptX(TheLayer, k + 1) THEN SwapControlPoints(k, k+1)
    IF (ABS(CptY(TheLayer, k) - CptY(TheLayer, k + 1)) > TheMaxChange) AND _
    IF (ABS(CptY(TheLayer, k) - CptY(TheLayer, k + 2)) > TheMaxChange) THEN
        CptY(TheLayer, k) = CptY(TheLayer, k + 1)
    END IF

    'Check adjacent points in middle
    FOR k = 1 TO CptN(TheLayer) -1
        IF CptX(TheLayer, k) > CptX(TheLayer, k + 1) THEN SwapControlPoints(k, k+1)

        IF (ABS(CptY(TheLayer, k) - CptY(TheLayer, k - 1)) > TheMaxChange) AND _
        IF (ABS(CptY(TheLayer, k) - CptY(TheLayer, k + 1)) > TheMaxChange) THEN
            CptY(TheLayer, k) = (CptY(TheLayer, k - 1) + CptY(TheLayer, k + 1)) \ 2
        END IF
    NEXT k

    'Check last point
    k = CptN(TheLayer)
    IF CptX(TheLayer, k) < CptX(TheLayer, k - 1) THEN SwapControlPoints(k, k-1)
    IF (ABS(CptY(TheLayer, k) - CptY(TheLayer, k - 1)) > TheMaxChange)  THEN
        CptY(TheLayer, k) = CptY(TheLayer, k - 1)
    END IF


END SUB



SUB SwapControlPoints(indx1 AS INTEGER, indx2 AS INTEGER)
    SWAP(CptX(CurrLayer, indx1), CptX(CurrLayer, indx2))
    SWAP(CptY(CurrLayer, indx1), CptY(CurrLayer, indx2))
    '...dont need to swap CtrlPt panels if you perform GetSplines ...
END SUB




SUB SplineMouseMove(X AS INTEGER, Y AS INTEGER, Shift AS INTEGER)
    IF DrawFlag THEN                        ' (SplineLocked(CurrLayer) = False)
        INC(SegN(CurrLayer), 1)
        IF SegN(CurrLayer) > MaxPts THEN SegN(CurrLayer) = MaxPts
        DEFINT n = SegN(CurrLayer)          'make it easier to read
        DEFINT j = FindMaxVertChange(SegmentBMP, X - OCT_Left, Y - 4, y + 4)            'FindMaxEdgeChange??
        SegX(CurrLayer, n) = x
        SegY(CurrLayer, n) = IIF(j <> NullResult, j, y)
        Form.Line(SegX(CurrLayer, n-1), SegY(CurrLayer, n-1), SegX(CurrLayer, n), SegY(CurrLayer, n), SegColor(CurrLayer))
    END IF
END SUB




SUB SplineMouseUp(button AS LONG , x AS LONG , y AS LONG , shift AS INTEGER)
    DrawFlag = False
    IF SplineLocked(CurrLayer) = False THEN
        Form.Cursor = crDefault
        IF SegN(CurrLayer) < 3 THEN                             'user error, short click and drag
            SegN(CurrLayer) = 0
            EXIT SUB 
        END IF

        SplineLocked(CurrLayer) = True                          'lock this layer from adding more data on subsequent click and drag
        DEFINT s = IIF(SegN(CurrLayer) < 20, 1, 5)              'set resolution of control points
        DEFINT k 
        FOR k = 0 TO SegN(CurrLayer)
            IF (k MOD s) = 0 THEN AddControlPoint(SegX(CurrLayer, k), SegY(CurrLayer, k))
        NEXT k

        'need at least 4 carets, 0 based
        IF CptN(CurrLayer) < 3 THEN
            FOR k = CptN(CurrLayer) TO 3
                AddControlPoint(x + k * 12 , y)
                INC(SegN(CurrLayer), 1)
                SegX(CurrLayer, k) = x + k * 12
                SegY(CurrLayer, k) = y
            NEXT k
        END IF 

        GetSpline
        ShowSplineControlPoints(True)
        DrawOCTSegmentation
    END IF
END SUB




SUB SplineMouseDown(button AS LONG , x AS LONG , y AS LONG , shift AS INTEGER)
    'first check  we start drawing a segmentation line
    IF SplineLocked(CurrLayer) = False THEN
        SegN(CurrLayer) = 0
        SegX(CurrLayer, 0) = x
        SegY(CurrLayer, 0) = y
        DrawFlag = True
        EXIT SUB
    END IF

    'then check if a segmentation line was clicked
    OCT_Layer_Picked = GetLayerIsPicked(X, Y)
    IF OCT_Layer_Picked = cNone THEN EXIT SUB
    CurrLayer = OCT_Layer_Picked
    IF SplineButton.Tag THEN ShowSplineControlPoints (True) 
END SUB



SUB SplineButton_Click
    DEFINT i, ShowEm
    IF SplineButton.Caption = "Spline" THEN     'switch to lines
        SplineButton.Tag = False
        SplineButton.Caption = "Lines"
        OCTThicknessLabel.Visible = True
        ShowSplineControlPoints (False)
        ShowEm = False
    ELSE                                        'switch to spline
        SplineButton.Tag = True
        SplineButton.Caption = "Spline"
        OCTThicknessLabel.Visible = False
        ShowEm = True
    END IF

    FOR i = 0 TO MaxLayers
        LayerShowBox(i).Visible = ShowEm
        LayerDelete(i).Visible = ShowEm
    NEXT i

    DrawOCTSegmentation
    FormRepaint
END SUB



SUB CtrlPtMouseDown(Button%, X%, Y%, Shift%, Sender AS QPANEL)
    CtrlPtDrag = True
    OrigX = Screen.MouseX
    OrigY = Screen.MouseY
    LastCtrlPt = Sender.Tag
END SUB


SUB CtrlPtMouseUp(Button%, X%, Y%, Shift%, Sender AS QPANEL)
    IF Button% = mouseRtButton THEN 
        DeleteControlPoint(sender.Tag)
        GetSpline
        DrawOCTSegmentation
        FormRepaint
    END IF
    CtrlPtDrag = False
END SUB



SUB CtrlPtMouseMove(x AS LONG , y AS LONG , shift AS INTEGER , sender AS QPANEL)
    STATIC ImBusy

    IF ImBusy = False THEN
        ImBusy = True

        IF DeleteKeyDown THEN
            DeleteControlPoint(sender.Tag)
            GetSpline
            DrawOCTSegmentation
            FormRepaint
            ImBusy = False
            EXIT SUB
        END IF


        IF CtrlPtDrag THEN
            DEFINT Dx = Screen.MouseX - OrigX
            DEFINT Dy = Screen.MouseY - OrigY

            IF Shift THEN           'user holds down shift key, move whole line
                DEFINT i
                FOR i = 0 TO CptN(CurrLayer)
                    CtrlPt(i).Left = CtrlPt(i).Left + Dx 
                    CtrlPt(i).Top  = CtrlPt(i).Top + Dy 
                    CptX(CurrLayer, i) = CptX(CurrLayer, i) + Dx
                    CptY(CurrLayer, i) = CptY(CurrLayer, i) + Dy
                NEXT i
            ELSE
                sender.Left = sender.Left + Dx
                sender.Top =  sender.Top  + Dy
                CptX(CurrLayer, sender.Tag) = sender.Left + CtrlPtHalfWidth
                CptY(CurrLayer, sender.Tag) = sender.Top  + CtrlPtHalfHeight
            END IF
            OrigX = Screen.MouseX 
            OrigY = Screen.MouseY
            DOEVENTS
            GetSpline
            DrawOCTSegmentation
            FormRepaint             'need this to update background
        END IF

        ImBusy = False
    END IF
END SUB



'---  run this sub if flag ShiftControlPointsKey is set ---
SUB MoveControlPointByCircle(x AS INTEGER, y AS INTEGER)

    STATIC ImBusy
    Form.Draw(x-CircleCursor.Width\2, y - CircleCursor.Height\2, CircleCursor.BMP)

    IF ImBusy = False THEN
        ImBusy = True

        DEFINT SphereSize = CircleCursor.Width\2
        DEFINT DistCriteria = SphereSize * 0.85 
        DEFINT i = 0
        DEFINT dx, dy, dist

        '------- get initial orientation on first entry  --------
        IF ControlPointCursorSGN = NullResult THEN
            DEFINT MinDist = 99999
            DO                                                              'search left to right
                dx = x - CtrlPt(i).Left - CtrlPtHalfHeight
                dy = y - CtrlPt(i).Top  - CtrlPtHalfHeight
                dist = SQR(dx * dx + dy * dy)                               'get attraction distance from cursor
                IF dist < MinDist THEN 
                    MinDist = dist
                    ControlPointCursorSGN = SGN(dy)
                END IF
                i ++
            LOOP UNTIL (i > CptN(CurrLayer))
            IF ControlPointCursorSGN = 0 THEN ControlPointCursorSGN = 1
        END IF

        '------- shift control points  --------
        i = 0
        DO                                                              'search left to right
            dx = x - CtrlPt(i).Left - CtrlPtHalfHeight
            dy = y - CtrlPt(i).Top  - CtrlPtHalfHeight
            dist = SQR(dx * dx + dy * dy)                               'get attraction distance from cursor

            IF (dist < DistCriteria) OR (ABS(dx) < (SphereSize/3)) THEN
                CtrlPt(i).Top = y - (SphereSize * ControlPointCursorSGN) - CtrlPtHalfHeight
                'CtrlPt(i).Top = y - (sqr(dist * dist - dx * dx) * ControlPointCursorSGN) - CtrlPtHalfHeight  'terrible
                CptY(CurrLayer, i) = CtrlPt(i).Top + CtrlPtHalfHeight
            END IF

            i ++
        LOOP UNTIL (i > CptN(CurrLayer))

        DOEVENTS
        GetSpline
        DrawOCTSegmentation
        FormRepaint             'need this to update background
    END IF

    ImBusy = False
END SUB



SUB DrawSplines
    DEFINT i, j
    FOR j = 0 TO MaxLayers
        
        IF (SegN(j) > 0) AND (LayerShowBox(j).Checked) THEN
            FOR i = 1 TO  SegN(j)
                TweakOCT.Line(SegX(j, i-1) - OCT_Left, SegY(j, i-1), SegX(j, i) - OCT_Left, SegY(j, i), SegColor(j))
            NEXT i
        END IF
    NEXT j
END SUB



'-------------------------------------------------------------------------------------------
'just do sorting left to right for final scoring of segmentation for search purposes 
'-------------------------------------------------------------------------------------------
SUB ReverseSpline
    DEFINT i = 0
    DO
        SwapControlPoints(i, CptN(CurrLayer) - i)
        'SWAP(CptX(CurrLayer, i), CptX(CurrLayer, CptN(CurrLayer) - i))
        'SWAP(CptY(CurrLayer, i), CptY(CurrLayer, CptN(CurrLayer) - i))
        'dont need to swap CtrlPt panels...
        i ++
        IF i >= (CptN(CurrLayer) - i) THEN EXIT DO
    LOOP UNTIL (i > CptN(CurrLayer)\2)              'just in case...
    GetSpline                                       're-calculate
END SUB




FUNCTION SplinePoint(P() AS INTEGER, t AS SINGLE) AS INTEGER
    RESULT = 0.5 * ((2.0 * P(1)) + _
        (- 1.0 * P(0) + P(2)) * t + _
        (2.0 * P(0) - 5.0 * P(1) + 4.0 * P(2) - P(3)) * t * t + _
        (- 1.0 * P(0) + 3.0 * P(1) - 3.0 * P(2) + P(3)) * t * t * t)
END FUNCTION




SUB GetSpline

    DEFINT Num = CptN(CurrLayer)                                                                'number of control points to work with
    IF Num < 3 THEN EXIT SUB                                                                    'got nothing

    DEFSNG t                                                                                    'location in between control points
    DEFINT n , k = 0
    DEFINT Px(3)
    DEFINT Py(3)
    DIM x(MaxCtrlPts) AS INTEGER 
    DIM y(MaxCtrlPts) AS INTEGER


    'copy the control point x, y locations
    MEMSET(VARPTR(x(0)), 0, (1 + MaxCtrlPts) * SIZEOF(INTEGER))                                 'clear static arrays!!
    MEMSET(VARPTR(y(0)), 0, (1 + MaxCtrlPts) * SIZEOF(INTEGER))                                 'clear static arrays!!

    MEMCPY(VARPTR(x(0)), VARPTR( CptX(CurrLayer, 0)), (MaxCtrlPts+1) * SIZEOF(INTEGER))         'copy x-axis control points of current slice to working array
    MEMCPY(VARPTR(y(0)), VARPTR( CptY(CurrLayer, 0)), (MaxCtrlPts+1) * SIZEOF(INTEGER))         'copy y-axis control points of current slice to working array

    'set up first 4 points with leftmost extrapolated
    Px(0) = x(0) + (10 * SGN(x(0) - x(1)))      'first set for spline is horizontal offset
    Px(1) = x(0)
    Px(2) = x(1)
    Px(3) = x(2)

    Py(0) = y(0)                                'first set for spline is horizontal offset
    Py(1) = y(0)
    Py(2) = y(1)
    Py(3) = y(2)

    DEFSNG stepsize     'resolution of spline points between control points

    stepsize = SplineGap / SQR( (Px(1) - Px(2)) * (Px(1) - Px(2)) + (Py(1) - Py(2)) * (Py(1) - Py(2)) )
    FOR t = 0 TO 1.0! STEP stepsize
        SegX(CurrLayer, k) = SplinePoint( Px(), t)
        SegY(CurrLayer, k) = SplinePoint( Py(), t)
        k ++
    NEXT t


    FOR n = 1 TO (Num - 2)
        Px(0) = x(n - 1)
        Px(1) = x(n)    
        Px(2) = x(n + 1)
        Px(3) = x(n + 2)

        Py(0) = y(n - 1)
        Py(1) = y(n)    
        Py(2) = y(n + 1)
        Py(3) = y(n + 2)

        stepsize = SplineGap / SQR( (Px(1) - Px(2)) * (Px(1) - Px(2)) + (Py(1) - Py(2)) * (Py(1) - Py(2)) )
        'set spline line x, y locations
        FOR t = 0 TO 1.0! STEP stepsize
            SegX(CurrLayer, k) = SplinePoint( Px(), t)
            SegY(CurrLayer, k) = SplinePoint( Py(), t)
            k ++
            IF k > MaxPts THEN k = MaxPts
        NEXT t
    NEXT n

    'set up last segment
    Px(0) = x(Num-2)                                                        'last set for spline is straight horizontal
    Px(1) = x(Num-1)
    Px(2) = x(Num)  
    Px(3) = x(Num) + 10 * SGN(x(Num) - x(Num-1))

    Py(0) = y(Num-2)
    Py(1) = y(Num-1)
    Py(2) = y(Num)  
    Py(3) = y(Num)  

    'set last spline line x, y locations
    stepsize = SplineGap / SQR( (Px(1) - Px(2)) * (Px(1) - Px(2)) + (Py(1) - Py(2)) * (Py(1) - Py(2)) )
    FOR t = 0 TO 1.0! STEP stepsize
        SegX(CurrLayer, k) = SplinePoint( Px(), t)
        SegY(CurrLayer, k) = SplinePoint( Py(), t)
        k ++
        IF k > MaxPts THEN k = MaxPts
    NEXT t

    k--
    SegN(CurrLayer) = k

END SUB



SUB ShowSplineControlPoints( VisFlag  AS INTEGER)
    DEFINT k, n

    IF VisFlag = False THEN

        IF CtrlPt(0).Visible THEN
            FOR k = 0 TO MaxCtrlPts
                CtrlPt(k).Visible = False
            NEXT k
        END IF

    ELSE

        n = CptN(CurrLayer)
        IF n > cNone THEN
            FOR k = 0 TO n
                CtrlPt(k).Left = CptX(CurrLayer, k) - CtrlPtHalfWidth
                CtrlPt(k).Top =  CptY(CurrLayer, k) - CtrlPtHalfHeight
                CtrlPt(k).Visible = True
            NEXT k
        END IF
        FOR k = (n+1) TO MaxCtrlPts         'hide extras
            CtrlPt(k).Visible = False
        NEXT k
    END IF
END SUB



SUB ClearLoci
    DEFINT i
    FOR i = 0 TO OCTLocus(0).Num -1                                         'how many do we have? won't plot if num = 0
        OCTLocus(i).ScreenX = 0
        OCTLocus(i).ScreenY = 0
        OCTLocus(i).ScreenThickness = 0
    NEXT i
    OCTLocus(0).Num = 0
    OCTThicknessLabel.Color = OCTPanel.Color
    OCTThicknessLabel.Caption = ""
END SUB



SUB LayerShowBox_Clicked (Sender AS QCHECKBOX)
    CurrLayer = Sender.Tag                                                  'set as current layer no matter what
    IF SplineButton.Tag THEN ShowSplineControlPoints (Sender.Checked)
    DrawOCTSegmentation                                                     'Draws splines
    FormRepaint
END SUB


SUB LayerDelete_Click(Sender AS QBUTTON)
    CurrLayer = Sender.Tag

    IF SplineButton.Tag THEN
        IF SegN(CurrLayer) = 0 THEN EXIT SUB
    ELSE
        IF OCTLocus(0).Num = 0 THEN EXIT SUB
    END IF

    IF MessageBox("Clear current segmentation then draw a new one?", "Clear segmentation", &H30 + 1) = 1 THEN 
        IF SplineButton.Tag THEN
            ShowSplineControlPoints (False)
            SegN(CurrLayer) = 0
            CptN(CurrLayer) = -1    'set to none
            SplineLocked(CurrLayer) = False
            Form.Cursor = crHandPoint
        ELSE
            DEFINT i
            FOR i = 0 TO OCTLocus(0).Num -1                                 'how many do we have? won't plot if num = 0
                OCTLocus(i).ScreenX = 0
                OCTLocus(i).ScreenY = 0
                OCTLocus(i).ScreenThickness = 0
            NEXT i
        END IF

        DrawOCTSegmentation
        FormRepaint
    END IF
END SUB


SUB LockAllSplines(TheFlag AS INTEGER)
    DrawFlag = False                                                        'prevent mouse drag & drawing
    DEFINT TheLayer
    FOR TheLayer = 0 TO MaxLayers
        SplineLocked(TheLayer) = TheFlag
    NEXT TheLayer
END SUB



SUB CalcAllLayerSplines
    DEFINT TheLayer
    FOR TheLayer = 0 TO MaxLayers
        CurrLayer = TheLayer
        GetSpline                                                           'calculate spline fit to segmentation
    NEXT TheLayer
END SUB


'----------------------------------------------------------
'  common SUB for segmentation and events
'----------------------------------------------------------

SUB DoSegmentation
    SegmentOCTimage                                                     'handles autoseg checkbox
    DrawOCTSegmentation                                                 'Draws splines
    FormRepaint
END SUB

