'------------------------------------------------------------------------------------------'
'    adapted from Linear Solver R.1 26Feb04                                                             '
'    Solves System of Linear Equations and does Polynomial Regression                      '
'    By: puddlduk@yahoo.com                                                                '
'
' -------------------------------------------------------------------------------------------------------'
'  Input an array of x and corresponding y values in double format, and # of elements
'  PolyOrder = order of polynomial fit (should be well under # of elements
'  Output an array of exponent coefficients in double format (size PolyOrder + 1) 
'    that can be used to predict y values, based on any value of x
' ------------------------------------------------------------------------------------------------------'


FUNCTION FitCurve (InputDataX() AS DOUBLE, InputDataY() AS DOUBLE, NumPts AS INTEGER, PolyOrder AS INTEGER, OutCoeff() AS DOUBLE ) AS INTEGER

    DEFINT Rtn = 0

    IF NumPts < (PolyOrder - 2) THEN
        RESULT = Rtn
        EXIT FUNCTION
    END IF

    DEFINT i , j , k , jmax , n , l , s
    DEFDBL T , coeff , sm , st , xpoint , ypoint , x , y
    DEFINT MatrixSize = 4

    n = PolyOrder + 1                        ' Matrix size
    MatrixSize = n

    DEFDBL MatrixIM(MatrixSize, MatrixSize+1)
    DEFDBL Coef(MatrixSize)
    DEFDBL Matrix(MatrixSize, MatrixSize)

    for j = 0 to PolyOrder
        for i = 0 to PolyOrder
            sm = 0                   ' Matrix cell summation
            for s = 0 to (NumPts-1)
                sm = sm + InputDataX(s)^(i+j)
            next s
            Matrix(i,j) = sm
        next i
        coeff = 0                    ' Coefficients cell summation
        for s = 0 to (NumPts-1)
            coeff = coeff + InputDataY(s) * InputDataX(s)^j
        next s
        Coef(j) = coeff
    next j

    ' ----------------------------------------------------------------------------------------------'
    ' Gauss Solver Routine                                                                          '
    ' n = polynomial order?
    ' ----------------------------------------------------------------------------------------------'

    n = MatrixSize -1 

    for i = 0 to n                                    'Populates augmented matrix 
        for j = 0 to n                                'i= column number, j= row number 
            MatrixIM(i,j) = Matrix(i,j)
        next j
        MatrixIM(n+1,i) = -(Coef(i))
        OutCoeff(i) = 0
    next i

    for j= 0 to n                  
        jmax = j
        T=abs( MatrixIM(jmax,j))   ' Search for largest element in this column
        for l = j+1 to n                                      
            if t < abs(MatrixIM(l,j)) then
                jmax = l
                t = abs(MatrixIM(l,jmax))
            end if 
        next l
        if j < jmax then                               ' Switch rows as needed     
            for k = j to n+1 
                st = MatrixIM(k,j)
                MatrixIM(k,j) = MatrixIM(k,jmax)
                MatrixIM(k,jmax) = st
            next k
        end if                      
        for l = j+1 to n                       ' construct echelon (through j loop)
            t = MatrixIM(j,l) / MatrixIM(j,j)
            MatrixIM(j,l)=0
            for i = j+1 to n+1
                MatrixIM(i,l) = MatrixIM(i,l) - t * MatrixIM(i,j)
            next i
        next  l
     next j
     For l = n to 0 step -1                    ' Back-substitution to find solution
         t = MatrixIM(n+1,l)
         for i = (l+1) to n
             t = t + MatrixIM(i,l) * OutCoeff(i)
         next i
         OutCoeff(l) = -t / MatrixIM(l,l)
     next l

    RESULT = n
END FUNCTION




FUNCTION GetYprediction(inX AS DOUBLE, PolyOrder AS INTEGER, InCoeff() AS DOUBLE) AS DOUBLE
    DEFDBL yPred = 0.000
    DEFINT i

    FOR i = 0 to PolyOrder
        yPred = yPred + InCoeff(i) * (inX^i)
    NEXT i
    RESULT = yPred
END FUNCTION


